#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import List, Optional, TYPE_CHECKING

import dateutil
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import BUCKET_SIZE, DATA_DRIFT_METRIC
from datarobot.models.api_object import APIObject
from datarobot.models.deployment.mixins import MonitoringDataQueryBuilderMixin
from datarobot.utils import from_api

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class Period(TypedDict, total=False):
        start: datetime
        end: datetime

    class Percentile(TypedDict, total=False):
        percent: float
        value: float

    class MeanProbability(TypedDict, total=False):
        class_name: str
        value: float

    class ClassDistribution(TypedDict, total=False):
        class_name: str
        count: int
        percent: float

    class GeoPoint(TypedDict):
        latitude: float
        longitude: float

    class PredictionsOverTimeBucket(TypedDict, total=False):
        period: Period
        model_id: str
        row_count: Optional[int]
        mean_predicted_value: Optional[int]
        percentiles: List[Percentile]
        mean_probabilities: List[MeanProbability]
        class_distribution: List[ClassDistribution]
        mean_geo_centroid: Optional[GeoPoint]


class TargetDrift(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment target drift information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve target drift metric
    period : dict
        the time period used to retrieve target drift metric
    metric : str
        the data drift metric
    target_name : str
        name of the target
    drift_score : float
        target drift score
    sample_size : int
        count of data points for comparison
    baseline_sample_size : int
        count of data points for baseline
    """

    _path = "deployments/{}/targetDrift/"
    _period = t.Dict(
        {
            t.Key("start"): String >> dateutil.parser.parse,
            t.Key("end"): String >> dateutil.parser.parse,
        }
    )
    _converter = t.Dict(
        {
            t.Key("period"): _period,
            t.Key("metric"): t.Or(t.Enum(*DATA_DRIFT_METRIC.ALL), t.Null),
            t.Key("model_id"): t.Or(String(), t.Null),
            t.Key("target_name"): t.Or(String(), t.Null),
            t.Key("drift_score"): t.Or(t.Float(), t.Null),
            t.Key("sample_size"): t.Or(Int(), t.Null),
            t.Key("baseline_sample_size"): t.Or(Int(), t.Null),
            t.Key("segment_attribute", optional=True): t.String(),
            t.Key("segment_value"): t.Or(String(allow_blank=True), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        period=None,
        metric=None,
        model_id=None,
        target_name=None,
        drift_score=None,
        sample_size=None,
        baseline_sample_size=None,
        segment_attribute=None,
        segment_value=None,
    ):
        self.period = period or {}
        self.metric = metric
        self.model_id = model_id
        self.target_name = target_name
        self.drift_score = drift_score
        self.sample_size = sample_size
        self.baseline_sample_size = baseline_sample_size
        self.segment_attribute = segment_attribute
        self.segment_value = segment_value

    def __repr__(self) -> str:
        return "{}({} | {} | {} - {})".format(
            self.__class__.__name__,
            self.model_id,
            self.target_name,
            self.period.get("start"),
            self.period.get("end"),
        )

    @classmethod
    def get(
        cls,
        deployment_id: str,
        model_id: Optional[str] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        metric: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> TargetDrift:
        """Retrieve target drift information over a certain time period.

        .. versionadded:: v2.21

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        metric : str
            (New in version v2.22) metric used to calculate the drift score
        segment_attribute : Optional[str]
            (New in Version v3.6) the segment attribute
        segment_value : Optional[str]
            (New in Version v3.6) the segment value
        Returns
        -------
        target_drift : TargetDrift
            the queried target drift information

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, TargetDrift
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            target_drift = TargetDrift.get(deployment.id)
            target_drift.period['end']
            >>>'2019-08-01 00:00:00+00:00'
            target_drift.drift_score
            >>>0.03423
            accuracy.target_name
            >>>'readmitted'
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        if metric:
            params["metric"] = metric
        data = cls._client.get(path, params=params).json()
        data = from_api(data, keep_null_keys=True)
        return cls.from_data(data)


class FeatureDrift(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment feature drift information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve feature drift metric
    period : dict
        the time period used to retrieve feature drift metric
    metric : str
        the data drift metric
    name : str
        name of the feature
    drift_score : float
        feature drift score
    sample_size : int
        count of data points for comparison
    baseline_sample_size : int
        count of data points for baseline
    """

    _path = "deployments/{}/featureDrift/"
    _period = t.Dict(
        {
            t.Key("start"): String >> dateutil.parser.parse,
            t.Key("end"): String >> dateutil.parser.parse,
        }
    )
    _converter = t.Dict(
        {
            t.Key("period"): _period,
            t.Key("metric"): t.Or(t.Enum(*DATA_DRIFT_METRIC.ALL), t.Null),
            t.Key("model_id"): t.Or(String(), t.Null),
            t.Key("name"): t.Or(String(), t.Null),
            t.Key("drift_score"): t.Or(t.Float(), t.Null),
            t.Key("feature_impact"): t.Or(t.Float(), t.Null),
            t.Key("sample_size"): t.Or(Int(), t.Null),
            t.Key("baseline_sample_size"): t.Or(Int(), t.Null),
            t.Key("segment_attribute", optional=True): t.String(),
            t.Key("segment_value", optional=True): t.Or(String(allow_blank=True), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        period=None,
        metric=None,
        model_id=None,
        name=None,
        drift_score=None,
        feature_impact=None,
        sample_size=None,
        baseline_sample_size=None,
        segment_attribute=None,
        segment_value=None,
    ):
        self.period = period or {}
        self.metric = metric
        self.model_id = model_id
        self.name = name
        self.drift_score = drift_score
        self.feature_impact = feature_impact
        self.sample_size = sample_size
        self.baseline_sample_size = baseline_sample_size
        self.segment_attribute = segment_attribute
        self.segment_value = segment_value

    def __repr__(self) -> str:
        return "{}({} | {} | {} - {})".format(
            self.__class__.__name__,
            self.model_id,
            self.name,
            self.period.get("start"),
            self.period.get("end"),
        )

    @classmethod
    def list(
        cls,
        deployment_id: str,
        model_id: Optional[str] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        metric: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> List[FeatureDrift]:
        """Retrieve drift information for deployment's features over a certain time period.

        .. versionadded:: v2.21

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        metric : str
            (New in version v2.22) metric used to calculate the drift score
        segment_attribute : Optional[str]
            (New in Version v3.6) the segment attribute
        segment_value : Optional[str]
            (New in Version v3.6) the segment value

        Returns
        -------
        feature_drift_data : [FeatureDrift]
            the queried feature drift information

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, TargetDrift
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            feature_drift = FeatureDrift.list(deployment.id)[0]
            feature_drift.period
            >>>'2019-08-01 00:00:00+00:00'
            feature_drift.drift_score
            >>>0.252
            feature_drift.name
            >>>'age'
        """

        url = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        if metric:
            params["metric"] = metric
        response_json = cls._client.get(url, params=params).json()
        response_json = from_api(response_json, keep_null_keys=True)

        period = response_json.get("period", {})
        metric = response_json.get("metric")
        model_id = response_json.get("model_id")

        def _from_data_item(item):
            item["period"] = period
            item["metric"] = metric
            item["model_id"] = model_id
            return cls.from_data(item)

        data = []
        for item in response_json["data"]:
            data.append(_from_data_item(item))
        while response_json["next"] is not None:
            response_json = cls._client.get(response_json["next"]).json()
            response_json = from_api(response_json, keep_null_keys=True)
            for item in response_json["data"]:
                data.append(_from_data_item(item))

        return data


class PredictionsOverTime(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment predictions over time information.

    Attributes
    ----------
    baselines : List
        target baseline for each model queried
    buckets : List
        predictions over time bucket for each model and bucket queried
    """

    _path = "deployments/{}/predictionsOverTime/"
    _period = t.Dict(
        {
            t.Key("start"): String >> dateutil.parser.parse,
            t.Key("end"): String >> dateutil.parser.parse,
        }
    )
    _converter = t.Dict(
        {
            t.Key("baselines"): t.List(t.Dict().allow_extra("*")),
            t.Key("buckets"): t.List(t.Dict({"period": _period}).allow_extra("*")),
        }
    )

    def __init__(
        self,
        baselines: Optional[List[PredictionsOverTimeBucket]] = None,
        buckets: Optional[List[PredictionsOverTimeBucket]] = None,
    ):
        self.baselines = baselines or []
        self.buckets = buckets or []

    @classmethod
    def get(
        cls,
        deployment_id: str,
        model_ids: Optional[List[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        bucket_size: Optional[BUCKET_SIZE] = None,
        target_classes: Optional[List[str]] = None,
        include_percentiles: Optional[bool] = False,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> PredictionsOverTime:
        """Retrieve information for deployment's prediction response over a certain time period.

        .. versionadded:: v3.2

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_ids : list[str]
            ID of models to retrieve prediction stats
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        bucket_size : BUCKET_SIZE
            time duration of each bucket
        target_classes : list[str]
            class names of target, only for deployments with multiclass target
        include_percentiles : bool
            if the returned data includes percentiles,
            only for a deployment with a binary and regression target
        segment_attribute : Optional[str]
            (New in Version v3.6) the segment attribute
        segment_value : Optional[str]
            (New in Version v3.6) the segment value

        Returns
        -------
        predictions_over_time : PredictionsOverTime
            the queried predictions over time information
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_ids,
            bucket_size=bucket_size,
            target_class=target_classes,
            include_percentiles=include_percentiles,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from __future__ import annotations

from typing import Any, Dict

import trafaret as t

from datarobot.enums import enum_to_list, ModerationGuardAction
from datarobot.models.api_object import APIObject
from datarobot.models.moderation.intervention import GuardInterventionCondition

moderation_configuration_without_id = t.Dict(
    {
        t.Key("guard_conditions"): t.List(
            t.Dict(
                {
                    t.Key("comparator"): t.String,
                    t.Key("comparand"): t.Or(t.Float, t.String, t.Bool, t.List(t.String)),
                }
            )
        ),
        t.Key("intervention"): t.Dict(
            {
                t.Key("action"): t.String,
                t.Key("message"): t.String,
            }
        ),
    }
).ignore_extra("*")

moderation_configuration_with_id = moderation_configuration_without_id + {"id": t.String}


intervention = t.Dict(
    {
        t.Key("action"): t.Enum(*enum_to_list(ModerationGuardAction)),
        t.Key("message"): t.String,
    }
).ignore_extra("*")


class Intervention(APIObject):
    """Intervention for playground moderation.

    Attributes
    ----------
    action : str
        The intervention strategy.
    message : str
        The intervention message to replace the prediction when a guard condition is satisfied.
    """

    _converter = intervention

    def __init__(self, action: str, message: str):
        self.action = action
        self.message = message

    def __repr__(self) -> str:
        return f"Playground Intervention(action={self.action}, message={self.message})"

    def to_dict(self) -> Dict[str, Any]:
        return {
            "action": self.action,
            "message": self.message,
        }


class ModerationConfigurationWithoutId(APIObject):
    """The moderation configuration for a metric in Playground.

    Attributes
    ----------
    guard_conditions : GuardInterventionCondition
        The guard conditions for the metric.
    intervention : Intervention
        The interventions for the guard.
    """

    _converter = moderation_configuration_without_id

    def __init__(self, guard_conditions: list[Dict[str, Any]], intervention: Dict[str, str]):
        self.guard_conditions = [
            GuardInterventionCondition.from_server_data(gc) for gc in guard_conditions
        ]
        self.intervention = Intervention.from_server_data(intervention)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "guard_conditions": [gc.to_dict() for gc in self.guard_conditions],
            "intervention": self.intervention.to_dict(),
        }



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate

evaluation_dataset_configuration_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String(allow_blank=True),
        t.Key("size"): t.Int,
        t.Key("rows_count"): t.Int,
        t.Key("use_case_id"): t.String,
        t.Key("playground_id", optional=True): t.Or(t.String, t.Null),
        t.Key("dataset_id"): t.String,
        t.Key("dataset_name"): t.String(allow_blank=True),
        t.Key("prompt_column_name"): t.String,
        t.Key("response_column_name", optional=True): t.Or(t.String, t.Null),
        t.Key("user_name"): t.String(allow_blank=True),
        t.Key("correctness_enabled", optional=True): t.Or(t.Bool, t.Null),
        t.Key("creation_user_id"): t.String,
        t.Key("creation_date"): t.String,
        t.Key("tenant_id"): t.String,
        t.Key("execution_status"): t.String,
        t.Key("error_message", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


class EvaluationDatasetConfiguration(APIObject):
    """
    An evaluation dataset configuration used to evaluate the performance of LLMs.

    Attributes
    ----------
    id : str
        The evaluation dataset configuration ID.
    name : str
        The name of the evaluation dataset configuration.
    size : int
        The size of the evaluation dataset (in bytes).
    rows_count : int
        The row count of the evaluation dataset.
    use_case_id : str
        The ID of the Use Case associated with the evaluation dataset configuration.
    playground_id : Optional[str]
        The ID of the playground associated with the evaluation dataset configuration.
    dataset_id : str
        The ID of the evaluation dataset.
    dataset_name : str
        The name of the evaluation dataset.
    prompt_column_name : str
        The name of the dataset column containing the prompt text.
    response_column_name : Optional[str]
        The name of the dataset column containing the response text.
    user_name : str
        The name of the user who created the evaluation dataset configuration.
    correctness_enabled : Optional[bool]
        Whether correctness is enabled for the evaluation dataset configuration.
    creation_user_id : str
        The ID of the user who created the evaluation dataset configuration.
    creation_date : str
        The creation date of the evaluation dataset configuration (ISO-8601 formatted).
    tenant_id : str
        The ID of the DataRobot tenant this evaluation dataset configuration belongs to.
    execution_status : str
        The execution status of the evaluation dataset configuration.
    error_message : Optional[str]
        The error message associated with the evaluation dataset configuration.
    """

    _path = "api/v2/genai/evaluationDatasetConfigurations"
    _converter = evaluation_dataset_configuration_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        size: int,
        rows_count: int,
        use_case_id: str,
        dataset_id: str,
        dataset_name: str,
        prompt_column_name: str,
        user_name: str,
        creation_user_id: str,
        creation_date: str,
        tenant_id: str,
        execution_status: str,
        playground_id: Optional[str] = None,
        response_column_name: Optional[str] = None,
        correctness_enabled: Optional[bool] = None,
        error_message: Optional[str] = None,
    ):
        self.id = id
        self.name = name
        self.size = size
        self.rows_count = rows_count
        self.use_case_id = use_case_id
        self.playground_id = playground_id
        self.dataset_id = dataset_id
        self.dataset_name = dataset_name
        self.prompt_column_name = prompt_column_name
        self.response_column_name = response_column_name
        self.user_name = user_name
        self.correctness_enabled = correctness_enabled
        self.creation_user_id = creation_user_id
        self.creation_date = creation_date
        self.tenant_id = tenant_id
        self.execution_status = execution_status
        self.error_message = error_message

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, dataset_id={self.dataset_id})"

    @classmethod
    def get(cls, id: str) -> EvaluationDatasetConfiguration:
        """Get an evaluation dataset configuration by ID.

        Parameters
        ----------
        id: str
            The evaluation dataset configuration ID to fetch.

        Returns
        -------
        evaluation_dataset_configuration: EvaluationDatasetConfiguration
            The evaluation dataset configuration.
        """
        url = f"{cls._client.domain}/{cls._path}/{id}/"
        response_data = cls._client.get(url)
        return cls.from_server_data(response_data.json())

    @classmethod
    def list(
        cls,
        use_case_id: str,
        playground_id: str,
        evaluation_dataset_configuration_id: Optional[str] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = 100,
        sort: Optional[str] = None,
        search: Optional[str] = None,
        correctness_only: Optional[bool] = False,
        completed_only: Optional[bool] = False,
    ) -> List[EvaluationDatasetConfiguration]:
        """List all evaluation dataset configurations for a Use Case.

        Parameters
        ----------
        use_case_id: str
            The ID of the Use Case that evaluation datasets are returned for.
        playground_id: str
            The ID of the playground that evaluation datasets are returned for. Default is None.
        evaluation_dataset_configuration_id: Optional[str]
            The ID of the evaluation dataset configuration to fetch. Default is None.
        offset: Optional[int]
            The offset to start fetching evaluation datasets from. Default is 0.
        limit: Optional[int]
            The maximum number of evaluation datasets to return. Default is 100.
        sort: Optional[str]
            The order of return for evaluation datasets. Default is None, which returns sorting
            by creation time.
        search: Optional[str]
            A search term that filters results so that only evaluation datasets with names
            matching the string are returned. Default is None.
        correctness_only: Optional[bool]
            Whether to return only completed datasets (particularly applicable to completion of generated
            synthetic datasets). Default is False.
        completed_only: Optional[bool]
            Whether to return only completed datasets. Default is False.

        Returns
        -------
        evaluation_dataset_configurations: List[EvaluationDatasetConfiguration]
            A list of evaluation dataset configurations.
        """
        params = {
            "use_case_id": use_case_id,
            "playground_id": playground_id,
            "evaluation_dataset_configuration_id": evaluation_dataset_configuration_id,
            "offset": offset,
            "limit": limit,
            "sort": sort,
            "search": search,
            "correctness_only": correctness_only,
            "completed_only": completed_only,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    @classmethod
    def create(
        cls,
        name: str,
        use_case_id: str,
        dataset_id: str,
        prompt_column_name: str,
        playground_id: str,
        is_synthetic_dataset: bool = False,
        response_column_name: Optional[str] = None,
    ) -> EvaluationDatasetConfiguration:
        """Create an evaluation dataset configuration for an existing dataset.

        Parameters
        ----------
        name: str
            The name of the evaluation dataset configuration.
        use_case_id: str
            The Use Case ID that the evaluation dataset configuration will be added to.
        dataset_id: str
            An ID, to add to the configuration, that identifies the evaluation dataset.
        playground_id: str
            The ID of the playground that the evaluation dataset configuration will be added to.
            Default is None.
        prompt_column_name: str
            The name of the prompt column in the dataset.
        response_column_name: str
            The name of the response column in the dataset.
        is_synthetic_dataset: bool
            Whether the evaluation dataset is synthetic.


        Returns
        -------
        evaluation_dataset_configuration : EvaluationDatasetConfiguration
            The created evaluation dataset configuration.

        """
        url = f"{cls._client.domain}/{cls._path}/"
        payload = {
            "name": name,
            "use_case_id": use_case_id,
            "dataset_id": dataset_id,
            "prompt_column_name": prompt_column_name,
            "response_column_name": response_column_name,
            "is_synthetic_data": is_synthetic_dataset,
            "playground_id": playground_id,
        }
        response = cls._client.post(url, data=payload)
        return cls.from_server_data(response.json())

    def update(
        self,
        name: Optional[str] = None,
        dataset_id: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        response_column_name: Optional[str] = None,
    ) -> EvaluationDatasetConfiguration:
        """Update the evaluation dataset configuration.

        Parameters
        ----------
        name: Optional[str]
            The name of the evaluation dataset configuration.
        dataset_id: Optional[str]
            The ID of the dataset used in this configuration.
        prompt_column_name : Optional[str]
            The name of the prompt column in the dataset.
        response_column_name : Optional[str]
            The name of the response column in the dataset.

        Returns
        -------
        evaluation_dataset_configuration : EvaluationDatasetConfiguration
            The updated evaluation dataset configuration.

        """
        payload = {
            "name": name,
            "dataset_id": dataset_id,
            "prompt_column_name": prompt_column_name,
            "response_column_name": response_column_name,
        }
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        response = self._client.patch(url, data=payload)
        return self.from_server_data(response.json())

    def delete(self) -> None:
        """Delete the evaluation dataset configuration.

        Returns
        -------
        None
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Dict, Iterable, List, Optional

from mypy_extensions import TypedDict
import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject

SingleFeatureImpactTrafaret = t.Dict(
    {
        t.Key("feature_impacts"): t.List(
            t.Dict(
                {
                    t.Key("impact_unnormalized"): t.Float,
                    t.Key("impact_normalized"): t.Float,
                    t.Key("redundant_with"): t.Or(String, t.Null),
                    t.Key("feature_name"): String,
                }
            ).ignore_extra("*")
        )
    }
)


class FeatureImpactData(TypedDict):
    feature_name: str
    impact_normalized: float
    impact_unnormalized: float
    redundant_with: Optional[str]


class FeatureImpactsData(TypedDict):
    feature_impacts: List[FeatureImpactData]
    ran_redundancy_detection: bool
    row_count: Optional[int]


class FeatureImpactServerData(TypedDict):
    id: str
    entity_id: str
    project_id: str
    source: str
    data_slice_id: Optional[str]
    data: FeatureImpactsData


class FeatureImpact(APIObject):
    """
    Feature Impact data for model.

    Notes
    -----
    ``SingleFeatureImpactTrafaret`` is a dict containing the following:

        * ``feature_name`` (string) Name of the feature.
        * ``impact_unnormalized`` (float) How much worse the error metric score is
        when making predictions on modified data.
        * ``impact_normalized`` (float) How much worse the error metric score is
        when making predictions on modified data (like impactUnnormalized),
        but normalized such that the highest value is 1.
        * ``redundant_with`` (string or None) Name of feature that has the highest correlation
        with this feature.

    Attributes
    ----------
    count: int
        Number of features for which Feature Impact was run.
    ran_redundancy_detection: bool
        Indicates whether redundant feature identification was run while calculating Feature Impact.
    row_count: Optional[int]
        Number of rows used to calculate Feature Impact.
    shap_based: bool
        Whether SHAP impact was used to compute this Feature Impact; if False, permutation impact was used.
    data_slice_id: string, optional
        Slice to retrieve Feature Impact for; if None, retrieve unsliced data.
    backtest: Optional[int]
        Backtest of the record.
    """

    _converter = (
        t.Dict(
            {
                t.Key("count"): t.Int,
                t.Key("ran_redundancy_detection"): t.Bool,
                t.Key("row_count", optional=True, default=None): t.Or(t.Int, t.Null),
                t.Key("shap_based"): t.Bool,
                # to make newer client work with older DataRobot responses
                t.Key("backtest", optional=True, default=None): t.Or(
                    t.Int(gte=0), t.Null, t.Atom("holdout")
                ),
                t.Key("data_slice_id", optional=True, default=None): t.Or(String, t.Null),
            }
        )
        .merge(SingleFeatureImpactTrafaret)
        .ignore_extra("*")
    )

    def __init__(
        self,
        feature_impacts: List[FeatureImpactData],
        count: int,
        ran_redundancy_detection: bool,
        shap_based: bool,
        row_count: Optional[int] = None,
        backtest: Optional[int] = None,
        data_slice_id: Optional[str] = None,
    ):
        self.feature_impacts = feature_impacts
        self.count = count
        self.ran_redundancy_detection = ran_redundancy_detection
        self.row_count = row_count
        self.shap_based = shap_based
        self.backtest = backtest
        self.data_slice_id = data_slice_id

    @staticmethod
    def _repack_insights_response(server_data: Dict[str, Any]) -> Dict[str, Any]:
        """Repack the JSON sent by the GET /insights/ endpoint
        to match the format expected by the insight APIObject class.

        Parameters
        ----------
        server_data : dict
        {
           "id": "64957e82313c68964f8ffb60",
           "entityId": "649169eda7db5185a87674ef",
           "projectId": "647471b6b5d9cbd454f2cf63",
           "source": "training",
           "dataSliceId": None,
           "data": {
              "featureImpacts": [
                 {
                    "featureName": "readmitted",
                    "impactNormalized": 1.0,
                    "impactUnnormalized": 16.468279209120773,
                    "redundantWith": None,
                 },
              ],
              "ranRedundancyDetection": True,
              "rowCount": 2500,
           }
        }

        Returns
        -------
        dict
        {
           "count": 1,
           "featureImpacts": [
              {
                 "featureName": "readmitted",
                 "impactNormalized": 1.0,
                 "impactUnnormalized": 16.468279209120773,
                 "redundantWith": None,
              },
           ],
           "ranRedundancyDetection": True,
           "shapBased": False,
           "rowCount": 2500,
           "backtest": None,
           "dataSliceId": None,
        }
        """
        feature_impacts = server_data["data"]["featureImpacts"]
        return {
            "count": len(feature_impacts),
            "featureImpacts": feature_impacts,
            "ranRedundancyDetection": server_data["data"]["ranRedundancyDetection"],
            "shapBased": False,  # SHAP is not currently supported by /insights/
            "rowCount": server_data["data"]["rowCount"],
            "backtest": server_data["data"].get("backtest"),
            "dataSliceId": server_data["dataSliceId"],
        }

    @classmethod
    def from_server_data(  # type: ignore[override,no-untyped-def]
        cls,
        data: Dict[str, Any],
        keep_attrs: Optional[Iterable[str]] = None,
        use_insights_format: Optional[bool] = False,
        **kwargs,
    ):
        """
        Overwrite APIObject.from_server_data to handle feature impact data retrieved
        from either legacy URL or /insights/ new URL.

        Parameters
        ----------
        data : dict
            Directly translated dict of JSON from the server. No casing fixes have
            taken place.
        use_insights_format : Optional[bool]
            Whether to repack the data from the format used in the GET /insights/featureImpact/ URL
            to the format used in the legacy URL.
        """
        if use_insights_format:
            data = cls._repack_insights_response(data)

        return super().from_server_data(
            data=data,
            keep_attrs=["feature_impacts.redundant_with", "backtest", "row_count", "data_slice_id"],
        )



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.genai.playground_moderation_configuration import (
    moderation_configuration_without_id,
    ModerationConfigurationWithoutId,
)


def _get_obj(value: Union[Dict[str, Any], None], obj: Any) -> Any:
    return obj.from_server_data(value) if value is not None else None


ootb_metric_configuration_response_trafaret = t.Dict(
    {
        t.Key("ootb_metric_configuration_id"): t.String,
        t.Key("ootb_metric_name"): t.String,
        t.Key("execution_status"): t.String,
        t.Key("custom_ootb_metric_name", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_id", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_model_llm_validation_id", optional=True): t.Or(t.String, t.Null),
        t.Key("moderation_configuration", optional=True): t.Or(
            moderation_configuration_without_id, t.Null
        ),
        t.Key("error_message", optional=True): t.Or(t.String(), t.Null),
        t.Key("error_resolution", optional=True): t.Or(t.List(t.String()), t.Null),
    }
).ignore_extra("*")

ootb_metric_configuration_request_trafaret = t.Dict(
    {
        t.Key("ootb_metric_name"): t.String,
        t.Key("custom_ootb_metric_name", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_id", optional=True): t.Or(t.String, t.Null),
        t.Key("custom_model_llm_validation_id", optional=True): t.Or(t.String, t.Null),
        t.Key("moderation_configuration", optional=True): t.Or(
            moderation_configuration_without_id, t.Null
        ),
    }
).ignore_extra("*")

ootb_metric_configurations_trafaret = t.Dict(
    {
        t.Key("ootb_metric_configurations"): t.List(ootb_metric_configuration_response_trafaret),
    }
).ignore_extra("*")


class OOTBMetricConfigurationRequest(APIObject):
    """
    An object that represents a request for an out-of-the-box (OOTB) metric.

    Attributes
    ----------
    ootb_metric_name : str
        The DataRobot-defined name of the OOTB metric.
    custom_ootb_metric_name : Optional[str]
        The custom OOTB metric name chosen by the user.
    llm_id : Optional[str]
        The ID of the LLM to use for `correctness` and `faithfulness` metrics.
    custom_model_llm_validation_id : Optional[str]
        The ID of the custom model LLM validation (if using a custom model LLM).
    moderation_configuration : Optional[ModerationConfigurationWithoutId]
        The moderation configuration to be associated with the OOTB metric.
    """

    _converter = ootb_metric_configuration_request_trafaret

    def __init__(
        self,
        ootb_metric_name: str,
        custom_ootb_metric_name: Optional[str] = None,
        llm_id: Optional[str] = None,
        custom_model_llm_validation_id: Optional[str] = None,
        moderation_configuration: Optional[Dict[str, Any]] = None,
    ):

        self.ootb_metric_name = ootb_metric_name
        self.custom_ootb_metric_name = custom_ootb_metric_name
        self.llm_id = llm_id
        self.custom_model_llm_validation_id = custom_model_llm_validation_id
        self.moderation_configuration = _get_obj(
            moderation_configuration, ModerationConfigurationWithoutId
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(ootb_metric_name={self.ootb_metric_name})"

    def to_dict(self) -> Dict[str, Any]:
        return {
            "ootb_metric_name": self.ootb_metric_name,
            "custom_ootb_metric_name": self.custom_ootb_metric_name,
            "llm_id": self.llm_id,
            "custom_model_llm_validation_id": self.custom_model_llm_validation_id,
            "moderation_configuration": (
                self.moderation_configuration.to_dict()
                if self.moderation_configuration is not None
                else None
            ),
        }


class OOTBMetricConfigurationResponse(APIObject):
    """
    An object that represents a single OOTB metric.

    Attributes
    ----------
    ootb_metric_configuration_id : str
        The OOTB metric configuration ID.
    ootb_metric_name : str
        The DataRobot-defined name of the OOTB metric.
    execution_status : str
        The execution status of the OOTB metric.
    custom_ootb_metric_name : Optional[str]
        The custom OOTB metric name chosen by the user.
    llm_id : Optional[str]
        The ID of the LLM to use for `correctness` and `faithfulness` metrics.
    custom_model_llm_validation_id : Optional[str]
        The ID of the custom model LLM validation (if using a custom model LLM).
    moderation_configuration : Optional[ModerationConfigurationWithoutId]
        The moderation configuration to be associated with the OOTB metric.
    error_message : Optional[str]
        The error message associated with the OOTB metric configuration.
    error_resolution: Optional[list[str]]
        The error type associated with the insight error status.
    """

    _converter = ootb_metric_configuration_response_trafaret
    _path = "api/v2/genai/ootbMetricConfigurations"

    def __init__(
        self,
        ootb_metric_configuration_id: str,
        ootb_metric_name: str,
        execution_status: str,
        custom_ootb_metric_name: Optional[str] = None,
        llm_id: Optional[str] = None,
        custom_model_llm_validation_id: Optional[str] = None,
        moderation_configuration: Optional[Dict[str, Any]] = None,
        error_message: Optional[str] = None,
        error_resolution: Optional[List[str]] = None,
    ):
        self.ootb_metric_configuration_id = ootb_metric_configuration_id
        self.ootb_metric_name = ootb_metric_name
        self.execution_status = execution_status
        self.custom_ootb_metric_name = custom_ootb_metric_name
        self.llm_id = llm_id
        self.custom_model_llm_validation_id = custom_model_llm_validation_id
        self.moderation_configuration = _get_obj(
            moderation_configuration, ModerationConfigurationWithoutId
        )
        self.error_message = error_message
        self.error_resolution = error_resolution

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(ootb_metric_name={self.ootb_metric_name})"

    @classmethod
    def get(cls, ootb_metric_configuration_id: str) -> OOTBMetricConfigurationResponse:
        """Get the OOTB metric configuration."""
        response_data = cls._client.get(
            f"{cls._client.domain}/{cls._path}/{ootb_metric_configuration_id}/"
        )
        return cls.from_server_data(response_data.json())

    def delete(self) -> None:
        """Delete the OOTB metric configuration."""
        url = f"{self._client.domain}/{self._path}/{self.ootb_metric_configuration_id}/"
        self._client.delete(url)


class PlaygroundOOTBMetricConfiguration(APIObject):
    """OOTB metric configurations for a playground.

    Attributes
    ----------
    ootb_metric_configurations: (List[OOTBMetricConfigurationResponse]): The list of the OOTB metric configurations.
    """

    _converter = ootb_metric_configurations_trafaret
    path = "api/v2/genai/playgrounds/{playground_id}/ootbMetricConfigurations"

    def __init__(
        self,
        ootb_metric_configurations: List[Dict[str, Any]],
    ):
        self.ootb_metric_configurations = [
            OOTBMetricConfigurationResponse.from_server_data(config)
            for config in ootb_metric_configurations
        ]

    @classmethod
    def get(cls, playground_id: str) -> PlaygroundOOTBMetricConfiguration:
        """Get OOTB metric configurations for the playground."""
        response_data = cls._client.get(
            f"{cls._client.domain}/{cls.path.format(playground_id=playground_id)}/"
        )
        return cls.from_server_data(response_data.json())

    @classmethod
    def create(
        cls, playground_id: str, ootb_metric_configurations: list[OOTBMetricConfigurationRequest]
    ) -> PlaygroundOOTBMetricConfiguration:
        """Create a new OOTB metric configurations."""
        payload = {
            "ootb_metric_configurations": [
                config.to_dict() for config in ootb_metric_configurations
            ],
        }
        url = f"{cls._client.domain}/{cls.path.format(playground_id=playground_id)}/"
        response_data = cls._client.post(url, data=payload)
        return cls.from_server_data(response_data.json())



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject

llm_cost_configuration = t.Dict(
    {
        t.Key("input_token_price"): t.Float,
        t.Key("reference_input_token_count"): t.Int,
        t.Key("output_token_price"): t.Float,
        t.Key("reference_output_token_count"): t.Int,
        t.Key("currency_code"): t.String,
        t.Key("llm_id"): t.String,
        t.Key("custom_model_llm_validation_id", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


cost_metric_configuration = t.Dict(
    {
        t.Key("cost_configuration_id"): t.String,
        t.Key("use_case_id"): t.String,
        t.Key("playground_id"): t.String,
        t.Key("cost_metric_configurations"): t.List(llm_cost_configuration),
        # New configurations require a name, old ones might not have it
        t.Key("name", optional=True): t.Or(t.String(allow_blank=True), t.Null),
    }
).ignore_extra("*")


class LLMCostConfiguration(APIObject):
    """Cost configuration for a specific LLM model; used for cost metric calculation.
    Price-per-token is price/reference token count.

    Attributes
    ----------
    input_token_price (float): The price of the input token.
    reference_input_token_count (int): The reference input token count.
    output_token_price (float): The price of the output token.
    reference_output_token_count (int): The reference output token count.
    currency_code (str): The currency code.
    llm_id (str): The LLM ID.
    custom_model_llm_validation_id (Optional[str]): The custom model LLM validation ID if llm_id is custom-model.

    """

    _converter = llm_cost_configuration

    def __init__(
        self,
        input_token_price: float,
        reference_input_token_count: int,
        output_token_price: float,
        reference_output_token_count: int,
        currency_code: str,
        llm_id: str,
        custom_model_llm_validation_id: Optional[str] = None,
    ):
        self.input_token_price = input_token_price
        self.reference_input_token_count = reference_input_token_count
        self.output_token_price = output_token_price
        self.reference_output_token_count = reference_output_token_count
        self.currency_code = currency_code
        self.llm_id = llm_id
        self.custom_model_llm_validation_id = custom_model_llm_validation_id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.llm_id})"

    def to_dict(self) -> Dict[str, Any]:
        return {
            "input_token_price": self.input_token_price,
            "reference_input_token_count": self.reference_input_token_count,
            "output_token_price": self.output_token_price,
            "reference_output_token_count": self.reference_output_token_count,
            "currency_code": self.currency_code,
            "llm_id": self.llm_id,
            "custom_model_llm_validation_id": self.custom_model_llm_validation_id,
        }


class CostMetricConfiguration(APIObject):
    """Cost metric configuration for a use case.

    Attributes
    ----------
    cost_configuration_id (str): The cost configuration ID.
    use_case_id (str): The use case ID.
    cost_metric_configurations (List[LLMCostConfiguration]): The list of LLM cost configurations.
    """

    _path = "api/v2/genai/costMetricConfigurations"
    _converter = cost_metric_configuration

    def __init__(
        self,
        cost_configuration_id: str,
        playground_id: str,
        use_case_id: str,
        cost_metric_configurations: List[Dict[str, Any]],
        name: str,
    ):
        self.cost_configuration_id = cost_configuration_id
        self.use_case_id = use_case_id
        self.cost_metric_configurations = [
            LLMCostConfiguration.from_server_data(config) for config in cost_metric_configurations
        ]
        self.playground_id = playground_id
        self.name = name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.cost_configuration_id})"

    @classmethod
    def get(cls, cost_metric_configuration_id: str) -> CostMetricConfiguration:
        """Get cost metric configuration by ID."""
        response_data = cls._client.get(
            f"{cls._client.domain}/{cls._path}/{cost_metric_configuration_id}/"
        )
        return cls.from_server_data(response_data.json())

    def update(
        self, cost_metric_configurations: List[LLMCostConfiguration], name: Union[str, None] = None
    ) -> CostMetricConfiguration:
        """Update the cost configurations."""
        payload: dict[str, Any] = {
            "cost_metric_configurations": [
                config.to_dict() for config in cost_metric_configurations
            ],
        }
        if name:
            payload["name"] = name
        url = f"{self._client.domain}/{self._path}/{self.cost_configuration_id}/"
        r_data = self._client.patch(url, data=payload)
        return self.from_server_data(r_data.json())

    @classmethod
    def create(
        cls,
        use_case_id: str,
        playground_id: str,
        name: str,
        cost_metric_configurations: List[LLMCostConfiguration],
    ) -> CostMetricConfiguration:
        """Create a new cost metric configuration."""
        payload = {
            "use_case_id": use_case_id,
            "cost_metric_configurations": [
                config.to_dict() for config in cost_metric_configurations
            ],
            "name": name,
            "playground_id": playground_id,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    def delete(self) -> None:
        """Delete the cost metric configuration."""
        url = f"{self._client.domain}/{self._path}/{self.cost_configuration_id}/"
        self._client.delete(url)



#
# Copyright 2021-2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional, Set, Union

import trafaret as t

from datarobot._compat import String
from datarobot.enums import KeyValueCategory, KeyValueEntityType, KeyValueType
from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate


class KeyValue(APIObject):
    """A DataRobot Key-Value.

    .. versionadded:: v3.4

    Attributes
    ----------
    id: str
        ID of the Key-Value
    created_at: str
        creation time of the Key-Value
    entity_id: str
        ID of the related Entity
    entity_type: KeyValueEntityType
        type of the related Entity
    name: str
        Key-Value name
    value: str
        Key-Value value
    numeric_value: float
        Key-Value numeric value
    boolean_value: bool
        Key-Value boolean value
    value_type: KeyValueType
        Key-Value type
    description: str
        Key-Value description
    creator_id: str
        ID of the user who created the Key-Value
    creator_name: str
        ID of the user who created the Key-Value
    category: KeyValueCategory
        Key-Value category
    artifact_size: int
        size in bytes of associated image, if applicable
    original_file_name: str
        name of uploaded original image or dataset file
    is_editable: bool
        true if a user with permissions can edit or delete
    is_dataset_missing: bool
        true if the key-value type is "dataset" and its dataset is not visible to the user
    error_message: str
        additional information if "isDataSetMissing" is true. Blank if there are no errors
    """

    _path = "keyValues/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("created_at"): String(),
            t.Key("entity_id"): String(),
            t.Key("entity_type"): t.Enum(*[e.value for e in KeyValueEntityType]),
            t.Key("name"): String(),
            t.Key("value"): String(),
            t.Key("numeric_value"): t.Float(),
            t.Key("boolean_value", optional=True, default=False): t.Bool(),
            t.Key("value_type"): t.Enum(*[e.value for e in KeyValueType]),
            t.Key("description"): String(allow_blank=True),
            t.Key("creator_id"): String(),
            t.Key("creator_name"): String(),
            t.Key("category"): t.Enum(*[e.value for e in KeyValueCategory]),
            t.Key("artifact_size"): t.Int(),
            t.Key("original_file_name"): String(allow_blank=True),
            t.Key("is_editable"): t.Bool(),
            t.Key("is_dataset_missing"): t.Bool(),
            t.Key("error_message"): String(allow_blank=True),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id: str,
        created_at: str,
        entity_id: str,
        entity_type: KeyValueEntityType,
        name: str,
        value: str,
        numeric_value: float,
        boolean_value: bool,
        value_type: KeyValueType,
        description: str,
        creator_id: str,
        creator_name: str,
        category: KeyValueCategory,
        artifact_size: int,
        original_file_name: str,
        is_editable: bool,
        is_dataset_missing: bool,
        error_message: str,
    ) -> None:
        self.id = id
        self.created_at = created_at
        self.entity_id = entity_id
        self.entity_type = KeyValueEntityType(entity_type)
        self.name = name
        self.value = value
        self.numeric_value = numeric_value
        self.boolean_value = boolean_value
        self.value_type = KeyValueType(value_type)
        self.description = description
        self.creator_id = creator_id
        self.creator_name = creator_name
        self.category = KeyValueCategory(category)
        self.artifact_size = artifact_size
        self.original_file_name = original_file_name
        self.is_editable = is_editable
        self.is_dataset_missing = is_dataset_missing
        self.error_message = error_message

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name or self.id!r})"

    def _update_values(self, new_response: KeyValue) -> None:
        fields: Set[str] = self._fields()  # type: ignore[no-untyped-call]
        for attr in fields:
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def _key_value_path(cls, key_value_id: str) -> str:
        return f"{cls._path}{key_value_id}/"

    @classmethod
    def get(cls, key_value_id: str) -> KeyValue:
        """Get Key-Value by id.

        .. versionadded:: v3.4

        Parameters
        ----------
        key_value_id: str
            ID of the Key-Value

        Returns
        -------
        KeyValue
            retrieved Key-Value

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = cls._key_value_path(key_value_id)
        return cls.from_location(path)

    @classmethod
    def list(cls, entity_id: str, entity_type: KeyValueEntityType) -> List[KeyValue]:
        """List Key-Values.

        .. versionadded:: v3.4

        Parameters
        ----------
        entity_id: str
            ID of the related Entity
        entity_type: KeyValueEntityType
            type of the related Entity

        Returns
        -------
        List[KeyValue]
            a list of Key-Values

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            cls._path,
            {"entityId": entity_id, "entityType": entity_type.value},
            cls._client,
        )
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def find(cls, entity_id: str, entity_type: KeyValueEntityType, name: str) -> Optional[KeyValue]:
        """Find Key-Value by name.

        .. versionadded:: v3.4

        Parameters
        ----------
        entity_id: str
            ID of the related Entity
        entity_type: KeyValueEntityType
            type of the related Entity
        name: str
            name of the Key-Value

        Returns
        -------
        List[KeyValue]
            a list of Key-Values

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            cls._path,
            {"entityId": entity_id, "entityType": entity_type.value, "name": name},
            cls._client,
        )
        kv_data = next(data, None)
        if not kv_data:
            return None
        return cls.from_server_data(kv_data)

    @classmethod
    def create(
        cls,
        entity_id: str,
        entity_type: KeyValueEntityType,
        name: str,
        category: KeyValueCategory,
        value_type: KeyValueType,
        value: Optional[Union[str, float, bool]] = None,
        description: Optional[str] = None,
    ) -> KeyValue:
        """Create a Key-Value.

        .. versionadded:: v3.4

        Parameters
        ----------
        entity_id: str
            ID of the associated resource
        entity_type: KeyValueEntityType
            type of the associated resource
        name: str
            name of the Key-Value. Cannot contain: { } ; |
        category: KeyValueCategory
            category of the Key-Value
        value_type: KeyValueType
            type of the Key-Value value
        value: Optional[Union[str, float, bool]]
            value of Key-Value
        description: Optional[str]
            description of the Key-Value

        Returns
        -------
        KeyValue
            created Key-Value

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        value_keys = {
            KeyValueType.NUMERIC: "numericValue",
            KeyValueType.BOOLEAN: "booleanValue",
        }
        value_key = value_keys.get(value_type, "value")

        response = cls._client.post(
            cls._path,
            data={
                "entityId": entity_id,
                "entityType": entity_type.value,
                "name": name,
                "category": category.value,
                "value_type": value_type.value,
                value_key: value,
                "description": description,
            },
        )

        return cls.get(response.json()["id"])

    def update(
        self,
        entity_id: Optional[str] = None,
        entity_type: Optional[KeyValueEntityType] = None,
        name: Optional[str] = None,
        category: Optional[KeyValueCategory] = None,
        value_type: Optional[KeyValueType] = None,
        value: Optional[Union[str, float, bool]] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
    ) -> None:
        """Update Key-Value.

        .. versionadded:: v3.4

        Parameters
        ----------
        entity_id: Optional[str]
            ID of the associated resource
        entity_type: Optional[KeyValueEntityType]
            type of the associated resource
        name: Optional[str]
            name of the Key-Value. Cannot contain: { } ; |
        category: Optional[KeyValueCategory]
            category of the Key-Value
        value_type: Optional[KeyValueType]
            type of the Key-Value value
        value: Optional[[Union[str, float, bool]]
            value of Key-Value
        description: Optional[str]
            description of the Key-Value
        comment: Optional[str]
            user comment explaining the change

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {
            "entityId": entity_id if entity_id is not None else self.entity_id,
            "entityType": entity_type.value if entity_type is not None else self.entity_type.value,
            "name": name if name is not None else self.name,
            "category": category.value if category is not None else self.category.value,
            "valueType": value_type.value if value_type is not None else self.value_type.value,
            "value": (
                self.value if value is None else None
            ),  # backend expects `value` to be always provided
        }
        if value is not None:
            value_keys = {
                KeyValueType.NUMERIC: "numericValue",
                KeyValueType.BOOLEAN: "booleanValue",
            }
            value_key = value_keys.get(value_type if value_type else self.value_type, "value")
            payload[value_key] = value  # type: ignore[assignment]
        if description is not None:
            payload["description"] = description
        if comment is not None:
            payload["comment"] = comment

        response = self._client.patch(self._key_value_path(self.id), data=payload)

        data = response.json()
        new_version = self.from_server_data(data)
        self._update_values(new_version)

    def refresh(self) -> None:
        """Update Key-Value with the latest data from server.

        .. versionadded:: v3.4

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.get(self.id)
        self._update_values(new_object)

    def delete(self) -> None:
        """Delete Key-Value.

        .. versionadded:: v3.4

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        path = self._key_value_path(self.id)
        self._client.delete(path)

    def get_value(self) -> Union[str, float, bool]:
        """Get a value of Key-Value.

        .. versionadded:: v3.4

        Returns
        -------
        Union[str, float, boolean]
            value depending on the value type
        """
        if self.value_type == KeyValueType.NUMERIC:
            return self.numeric_value
        elif self.value_type == KeyValueType.BOOLEAN:
            return self.boolean_value
        else:
            return self.value



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict

import trafaret as t

from datarobot.models.api_object import APIObject


class GuardNemoInfo(APIObject):
    """Details of a NeMo Guardrails moderation guard."""

    _converter = t.Dict(
        {
            t.Key("main_config"): t.String(),
            t.Key("rails_config"): t.String(),
            t.Key("blocked_terms"): t.String(),
            t.Key("llm_prompts"): t.String(),
            t.Key("actions"): t.String(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "NeMo guardrails; see details"

    def _set_values(
        self: GuardNemoInfo,
        main_config: str = "",
        rails_config: str = "",
        blocked_terms: str = "",
        llm_prompts: str = "",
        actions: str = "",
    ) -> None:
        """
        Initialize a NeMo guardrails object based on values from init or server data.
        """
        self.main_config = main_config
        self.rails_config = rails_config
        self.blocked_terms = blocked_terms
        self.llm_prompts = llm_prompts
        self.actions = actions

    def to_dict(self: GuardNemoInfo) -> Dict[str, str]:
        return {
            "mainConfig": self.main_config,
            "railsConfig": self.rails_config,
            "blockedTerms": self.blocked_terms,
            "llmPrompts": self.llm_prompts,
            "actions": self.actions,
        }



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Set

import trafaret as t

from datarobot._compat import String
from datarobot.enums import (
    GuardType,
    ModerationGuardLlmType,
    ModerationGuardOotbType,
    ModerationGuardStage,
)
from datarobot.models.api_object import APIObject
from datarobot.models.moderation.intervention import GuardInterventionForTemplate
from datarobot.models.moderation.model_info import GuardModelInfo
from datarobot.models.moderation.nemo_info import GuardNemoInfo
from datarobot.models.moderation.utils import exclude_if
from datarobot.utils import parse_time
from datarobot.utils.pagination import unpaginate


class ModerationTemplate(APIObject):
    """A DataRobot Moderation Template.

    .. versionadded:: v3.6

    Attributes
    ----------
    id: str
        ID of the Template
    name: str
        Template name
    """

    _path = "guardTemplates/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("description"): String(),
            t.Key("type"): t.Enum(*[e.value for e in GuardType]),
            t.Key("ootb_type", optional=True): t.Enum(*[e.value for e in ModerationGuardOotbType]),
            t.Key("llm_type", optional=True): t.Enum(*[e.value for e in ModerationGuardLlmType]),
            t.Key("allowed_stages"): t.List(t.Enum(*[e.value for e in ModerationGuardStage])),
            t.Key("created_at", optional=True): t.Or(parse_time, t.Null),
            t.Key("creator_id", optional=True): String(allow_blank=True),
            t.Key("org_id", optional=True): String(allow_blank=True),
            t.Key("intervention", optional=True): GuardInterventionForTemplate.schema,
            t.Key("model_info", optional=True): GuardModelInfo.schema,
            t.Key("nemo_info", optional=True): GuardNemoInfo.schema,
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self: ModerationTemplate,
        id: str,
        name: str,
        type: GuardType,
        description: str,
        created_at: str,
        allowed_stages: list[ModerationGuardStage],
        ootb_type: Optional[str] = None,
        llm_type: Optional[str] = None,
        creator_id: Optional[str] = None,
        org_id: Optional[str] = None,
        intervention: Optional[GuardInterventionForTemplate | Dict[str, Any]] = None,
        model_info: Optional[GuardModelInfo | Dict[str, Any]] = None,
        nemo_info: Optional[GuardNemoInfo | Dict[str, Any]] = None,
    ) -> None:
        self.id = id
        self.name = name
        self.type = GuardType(type)
        self.description = description
        self.created_at = created_at
        self.allowed_stages = [ModerationGuardStage(x) for x in allowed_stages]
        self.creator_id = creator_id
        self.org_id = org_id
        self.nemo_info = nemo_info
        self.ootb_type = None
        if ootb_type:
            self.ootb_type = ModerationGuardOotbType(ootb_type)
        self.llm_type = None
        if llm_type:
            self.llm_type = ModerationGuardLlmType(llm_type)
        # the following may arrive as objects or as dicts
        if intervention is None:
            self.intervention = None
        else:
            self.intervention = GuardInterventionForTemplate.ensure_object(intervention)
        if model_info is None:
            self.model_info = None
        else:
            self.model_info = GuardModelInfo.ensure_object(model_info)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name or self.id!r})"

    def _update_values(self: ModerationTemplate, new_response: ModerationTemplate) -> None:
        # called by update() and refresh()
        fields: Set[str] = self._fields()  # type: ignore[no-untyped-call]
        for attr in fields:
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def _template_path(cls, template_id: str) -> str:
        return f"{cls._path}{template_id}/"

    @classmethod
    def get(cls, template_id: str) -> ModerationTemplate:
        """Get Template by id.

        .. versionadded:: v3.6

        Parameters
        ----------
        template_id: str
            ID of the Template

        Returns
        -------
        ModerationTemplate
            retrieved Template

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = cls._template_path(template_id)
        return cls.from_location(path)

    @classmethod
    def list(cls) -> List[ModerationTemplate]:
        """List Templates.

        .. versionadded:: v3.6

        Parameters
        ----------
        none yet

        Returns
        -------
        List[ModerationTemplate]
            a list of Templates

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            cls._path,
            {},
            cls._client,
        )
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def find(cls, name: str) -> Optional[ModerationTemplate]:
        """Find Template by name.

        .. versionadded:: v3.6

        Parameters
        ----------
        name: str
            name of the Template

        Returns
        -------
        List[ModerationTemplate]
            a list of Templates

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            cls._path,
            {"name": name},
            cls._client,
        )
        kv_data = next(data, None)
        if not kv_data:
            return None
        return cls.from_server_data(kv_data)

    @classmethod
    def create(
        cls,
        name: str,
        description: str,
        type: GuardType,
        allowed_stages: List[ModerationGuardStage],
        intervention: Optional[GuardInterventionForTemplate] = None,
        ootb_type: Optional[ModerationGuardOotbType] = None,
        llm_type: Optional[ModerationGuardLlmType] = None,
        model_info: Optional[GuardModelInfo] = None,
        nemo_info: Optional[GuardNemoInfo] = None,
    ) -> ModerationTemplate:
        """Create a Template.

        .. versionadded:: v3.6

        Parameters
        ----------
        name: str
            name of the template.
        description: str
            description of the template.
        type: GuardType
            type of the template.
        allowed_stages: List[ModerationGuardStage]
            the stages of moderation this guard is allowed to be used
        ootb_type: ModerationGuardOotbType
            for guards of type "ootb", the specific "Out of the Box" metric type.
        llm_type:
            the backing LLM this guard uses.
        nemo_info
            additional configuration for NeMo Guardrails guards.
        model_info
            additional configuration for guards using a deployed model.
        intervention
            the assessment conditions, and action the guard should take if conditions are met.

        Returns
        -------
        ModerationTemplate
            created Template

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {
            "name": name,
            "type": type.value,
            "description": description,
            "allowedStages": allowed_stages,
            "ootbType": ootb_type,
            "llmType": llm_type,
            "intervention": intervention,
            "parameters": [],  # required but deprecated
        }
        if intervention:
            payload["intervention"] = intervention
        if model_info:
            payload["modelInfo"] = model_info
        if nemo_info:
            payload["nemoInfo"] = model_info

        response = cls._client.post(
            cls._path,
            data=payload,
        )

        # todo: POST response should include entire object, but only returns ID.
        # todo: re-fetch until that's fixed, then use cls.from_server_data(response.json())
        return cls.get(response.json()["id"])

    def update(
        self: ModerationTemplate,
        name: Optional[str] = None,
        description: Optional[str] = None,
        type: Optional[GuardType] = None,
        allowed_stages: Optional[List[ModerationGuardStage]] = None,
        intervention: Optional[GuardInterventionForTemplate] = None,
        ootb_type: Optional[ModerationGuardOotbType] = None,
        llm_type: Optional[ModerationGuardLlmType] = None,
        model_info: Optional[GuardModelInfo] = None,
        nemo_info: Optional[GuardNemoInfo] = None,
    ) -> None:
        """Update Template. All fields are optional, and omitted fields are left unchanged.

        .. versionadded:: v3.6

        Parameters
        ----------
        name: str
            name of the template.
        description: str
            description of the template.
        type: GuardType
            type of the template.
        allowed_stages: List[ModerationGuardStage]
            the stages of moderation this guard is allowed to be used
        ootb_type: ModerationGuardOotbType
            for guards of type "ootb", the specific "Out of the Box" metric type.
        llm_type:
            the backing LLM this guard uses.
        nemo_info
            additional configuration for NeMo Guardrails guards.
        model_info
            additional configuration for guards using a deployed model.
        intervention
            the assessment conditions, and action the guard should take if conditions are met.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """

        payload = {}
        payload.update(exclude_if("name", name))
        payload.update(exclude_if("description", description))
        payload.update(exclude_if("type", type))
        payload.update(exclude_if("ootb_type", ootb_type))
        payload.update(exclude_if("llm_type", llm_type))
        if allowed_stages is not None:
            payload.update({"allowedStages": [x.value for x in self.allowed_stages]})
        if nemo_info is not None:
            payload.update({"nemoInfo": nemo_info.to_dict()})
        if model_info is not None:
            payload.update({"nemoInfo": model_info.to_dict()})
        if intervention is not None:
            payload.update({"intervention": intervention.to_dict()})

        response = self._client.patch(self._template_path(self.id), data=payload)
        data = response.json()
        new_version = self.from_server_data(data)
        self._update_values(new_version)

    def refresh(self: ModerationTemplate) -> None:
        """Update Template with the latest data from server.

        .. versionadded:: v3.6

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.get(self.id)
        self._update_values(new_object)

    def delete(self: ModerationTemplate) -> None:
        """Delete Template.

        .. versionadded:: v3.6

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        path = self._template_path(self.id)
        self._client.delete(path)



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate

sidecar_model_metric_validation_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String(allow_blank=True),
        t.Key("target_column_name", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("prompt_column_name", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("deployment_id"): t.String,
        t.Key("validation_status"): t.String,
        t.Key("model_id"): t.String,
        t.Key("deployment_access_data", optional=True, default=None): t.Or(
            t.Null,
            t.Dict(
                {
                    t.Key("prediction_api_url"): t.String,
                    t.Key("datarobot_key", optional=True, default=None): t.Or(t.Null, t.String),
                    t.Key("authorization_header"): t.String,
                    t.Key("input_type"): t.String,
                    t.Key("model_type"): t.String,
                }
            ).ignore_extra("*"),
        ),
        t.Key("tenant_id"): t.String,
        t.Key("user_id"): t.String,
        t.Key("creation_date"): t.String,
        t.Key("error_message", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("deployment_name", optional=True, default=None): t.Or(
            t.Null, t.String(allow_blank=True)
        ),
        t.Key("user_name", optional=True, default=None): t.Or(t.Null, t.String(allow_blank=True)),
        t.Key("use_case_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("prediction_timeout"): t.Int,
        t.Key("playground_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("citations_prefix_column_name", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("response_column_name", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("expected_response_column_name", optional=True, default=None): t.Or(t.Null, t.String),
    }
).ignore_extra("*")


class SidecarModelMetricValidation(APIObject):
    """A sidecar model metric validation for LLMs.

    Attributes
    ----------
    id : str
        The ID of the sidecar model metric validation.
    prompt_column_name : str
        The name of the prompt column for the sidecar model.
    deployment_id : str
        The ID of the deployment associated with the sidecar model.
    model_id : str
        The ID of the sidecar model.
    validation_status : str
        The status of the validation job.
    deployment_access_data : dict
        The data that will be used for accessing the deployment prediction server.
        This field is only available for deployments that pass validation.
        Dict fields are as follows:
        - prediction_api_url - The URL for the deployment prediction server.
        - datarobot_key - The first of two auth headers for the prediction server.
        - authorization_header - The second of two auth headers for the prediction server.
        - input_type - The input type the model expects, either JSON or CSV.
        - model_type - The target type of the deployed custom model.
    tenant_id : str
        The ID of the tenant that created the sidecar model metric validation.
    name : str
        The name of the sidecar model metric.
    creation_date : str
        The date the sidecar model metric validation was created.
    user_id : str
        The ID of the user that created the sidecar model metric validation.
    deployment_name : str
        The name of the deployment associated with the sidecar model.
    user_name : str
        The name of the user that created the sidecar model metric validation.
    use_case_id : str
        The ID of the Use Case associated with the sidecar model metric validation.
    prediction_timeout : int
        The timeout in seconds for the prediction API used in this sidecar model metric validation.
    error_message : str
        Additional information for the errored validation.
    citations_prefix_column_name : str
        The name of the prefix in the citations column for the sidecar model.
    response_column_name : str
        The name of the response column for the sidecar model.
    expected_response_column_name : str
        The name of the expected response column for the sidecar model.
    target_column_name : str
        The name of the target column for the sidecar model.
    """

    _path = "api/v2/genai/sidecarModelMetricValidations"

    _converter = sidecar_model_metric_validation_trafaret

    def __init__(
        self,
        id: str,
        playground_id: str,
        prompt_column_name: str,
        deployment_id: str,
        model_id: str,
        validation_status: str,
        deployment_access_data: Optional[Dict[str, Any]],
        tenant_id: str,
        name: str,
        creation_date: str,
        user_id: str,
        deployment_name: Optional[str],
        user_name: Optional[str],
        use_case_id: Optional[str],
        prediction_timeout: int,
        error_message: Optional[str],
        citations_prefix_column_name: Optional[str],
        response_column_name: Optional[str],
        target_column_name: Optional[str],
        expected_response_column_name: Optional[str],
    ):
        self.id = id
        self.playground_id = playground_id
        self.prompt_column_name = prompt_column_name
        self.deployment_id = deployment_id
        self.model_id = model_id
        self.validation_status = validation_status
        self.deployment_access_data = deployment_access_data
        self.tenant_id = tenant_id
        self.error_message = error_message
        self.name = name
        self.creation_date = creation_date
        self.user_id = user_id
        self.deployment_name = deployment_name
        self.user_name = user_name
        self.use_case_id = use_case_id
        self.prediction_timeout = prediction_timeout
        self.error_message = error_message
        self.citations_prefix_column_name = citations_prefix_column_name
        self.response_column_name = response_column_name
        self.target_column_name = target_column_name
        self.expected_response_column_name = expected_response_column_name

    @classmethod
    def create(
        cls,
        deployment_id: str,
        name: str,
        prediction_timeout: int,
        model_id: Optional[str] = None,
        use_case_id: Optional[str] = None,
        playground_id: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        response_column_name: Optional[str] = None,
        citation_prefix_column_name: Optional[str] = None,
        expected_response_column_name: Optional[str] = None,
    ) -> SidecarModelMetricValidation:
        """Create a sidecar model metric validation.

        Parameters
        ----------
        deployment_id : str
            The ID of the deployment to validate.
        name : str
            The name of the validation.
        prediction_timeout : int
            The timeout in seconds for the prediction API used in this validation.
        model_id : Optional[str]
            The ID of the model to validate.
        use_case_id : Optional[str]
            The ID of the Use Case associated with the validation.
        playground_id : Optional[str]
            The ID of the playground associated with the validation.
        prompt_column_name : Optional[str]
            The name of the prompt column for the sidecar model.
        target_column_name : Optional[str]
            The name of the target column for the sidecar model.
        response_column_name : Optional[str]
            The name of the response column for the sidecar model.
        citation_prefix_column_name : Optional[str]
            The name of the prefix for citations column for the sidecar model.
        expected_response_column_name : Optional[str]
            The name of the expected response column for the sidecar model.

        Returns
        -------
        SidecarModelMetricValidation
            The created sidecar model metric validation.
        """

        payload = {
            "deployment_id": deployment_id,
            "name": name,
            "prediction_timeout": prediction_timeout,
            "model_id": model_id,
            "use_case_id": use_case_id,
            "playground_id": playground_id,
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "response_column_name": response_column_name,
            "citation_prefix_column_name": citation_prefix_column_name,
            "expected_response_column_name": expected_response_column_name,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        use_case_ids: Optional[List[str]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        search: Optional[str] = None,
        sort: Optional[str] = None,
        completed_only: Optional[bool] = True,
        deployment_id: Optional[str] = None,
        model_id: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        citation_prefix_column_name: Optional[str] = None,
    ) -> List[SidecarModelMetricValidation]:
        """List sidecar model metric validations.

        Parameters
        ----------
        use_case_ids : List[str], optional
            The IDs of the use cases to filter by.
        offset : Optional[int]
            The number of records to skip.
        limit : Optional[int]
            The maximum number of records to return.
        search : Optional[str]
            The search string.
        sort : Optional[str]
            The sort order.
        completed_only : Optional[bool]
            Whether to return only completed validations.
        deployment_id : Optional[str]
            The ID of the deployment to filter by.
        model_id : Optional[str]
            The ID of the model to filter by.
        prompt_column_name : Optional[str]
            The name of the prompt column to filter by.
        target_column_name : Optional[str]
            The name of the target column to filter by.
        citation_prefix_column_name : Optional[str]
            The name of the prefix for citations column to filter by.

        Returns
        -------
        List[SidecarModelMetricValidation]
            The list of sidecar model metric validations.

        """
        params = {
            "use_case_id": use_case_ids,
            "offset": offset,
            "limit": limit,
            "search": search,
            "sort": sort,
            "completed_only": completed_only,
            "deployment_id": deployment_id,
            "model_id": model_id,
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "citation_prefix_column_name": citation_prefix_column_name,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        response_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in response_data]

    @classmethod
    def get(cls, validation_id: str) -> SidecarModelMetricValidation:
        """Get a sidecar model metric validation by ID.

        Parameters
        ----------
        validation_id : str
            The ID of the validation to get.

        Returns
        -------
        SidecarModelMetricValidation
            The sidecar model metric validation.
        """
        url = f"{cls._client.domain}/{cls._path}/{validation_id}/"
        response_data = cls._client.get(url)
        return cls.from_server_data(response_data.json())

    def revalidate(self) -> SidecarModelMetricValidation:
        """Revalidate the sidecar model metric validation.

        Returns
        -------
        SidecarModelMetricValidation
            The sidecar model metric validation.

        """
        url = f"{self._client.domain}/{self._path}/{self.id}/revalidate/"
        response_data = self._client.post(url)
        return self.from_server_data(response_data.json())

    def update(
        self,
        name: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        response_column_name: Optional[str] = None,
        expected_response_column_name: Optional[str] = None,
        citation_prefix_column_name: Optional[str] = None,
        deployment_id: Optional[str] = None,
        model_id: Optional[str] = None,
        prediction_timeout: Optional[int] = None,
    ) -> SidecarModelMetricValidation:
        """Update the sidecar model metric validation.

        Parameters
        ----------
        name : Optional[str]
            The name of the validation.
        prompt_column_name : Optional[str]
            The name of the prompt column for the sidecar model.
        target_column_name : Optional[str]
            The name of the target column for the sidecar model.
        response_column_name : Optional[str]
            The name of the response column for the sidecar model.
        expected_response_column_name : Optional[str]
            The name of the expected response column for the sidecar model.
        citation_prefix_column_name : Optional[str]
            The name of the prefix for citations column for the sidecar model.
        deployment_id : Optional[str]
            The ID of the deployment to validate.
        model_id : Optional[str]
            The ID of the model to validate.
        prediction_timeout : Optional[int]
            The timeout in seconds for the prediction API used in this validation.

        Returns
        -------
        SidecarModelMetricValidation
            The updated sidecar model metric validation.
        """
        payload = {
            "name": name,
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": deployment_id,
            "model_id": model_id,
            "prediction_timeout": prediction_timeout,
            "response_column_name": response_column_name,
            "citation_prefix_column_name": citation_prefix_column_name,
            "expected_response_column_name": expected_response_column_name,
        }
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        response = self._client.patch(url, data=payload)
        return self.from_server_data(response.json())

    def delete(self) -> None:
        """Delete the sidecar model metric validation."""
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.genai.playground import Playground
from datarobot.utils.pagination import unpaginate


def get_entity_id(entity: Union[ComparisonChat, Playground, str]) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : APIObject or str
        Specifies either the entity ID or the entity.

    Returns
    -------
    id : str
        The entity ID.
    """
    return entity if isinstance(entity, str) else entity.id


comparison_chat_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String(allow_blank=True),
        t.Key("playground_id"): t.String,
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
    }
).ignore_extra("*")


class ComparisonChat(APIObject):
    """
    Metadata for a DataRobot GenAI comparison chat.

    Attributes
    ----------
    id : str
        The comparison chat ID.
    name : str
        The comparison chat name.
    playground_id : str
        The ID of the playground associated with the comparison chat.
    creation_date : str
        The date when the comparison chat was created.
    creation_user_id : str
        The ID of the creating user.
    """

    _path = "api/v2/genai/comparisonChats"

    _converter = comparison_chat_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        playground_id: str,
        creation_date: str,
        creation_user_id: str,
    ):
        self.id = id
        self.name = name
        self.playground_id = playground_id
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"

    @classmethod
    def create(
        cls,
        name: str,
        playground: Union[Playground, str],
    ) -> ComparisonChat:
        """
        Creates a new comparison chat.

        Parameters
        ----------
        name : str
            The comparison chat name.
        playground : Playground or str
            The playground associated with the created comparison chat, either `Playground`
            or playground ID.

        Returns
        -------
        comparison_chat : ComparisonChat
            The created comparison chat.
        """
        payload = {
            "playground_id": get_entity_id(playground),
            "name": name,
        }

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, comparison_chat: Union[ComparisonChat, str]) -> ComparisonChat:
        """
        Retrieve a single comparison chat.

        Parameters
        ----------
        comparison_chat : ComparisonChat or str
            The comparison chat you want to retrieve. Accepts `ComparisonChat` or
            comparison chat ID.

        Returns
        -------
        comparison_chat : ComparisonChat
            The requested comparison chat.
        """
        url = f"{cls._client.domain}/{cls._path}/{get_entity_id(comparison_chat)}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        playground: Optional[Union[Playground, str]] = None,
        sort: Optional[str] = None,
    ) -> List[ComparisonChat]:
        """
        List all comparison chats available to the user. If the playground is specified,
        results are restricted to only those comparison chats associated with the playground.

        Parameters
        ----------
        playground : Optional[Union[Playground, str]], optional
            Returns only those comparison chats associated with a particular playground,
            specified by either the `Playground` or the playground ID.
        sort : Optional[str]
            The property to sort comparison chats by. Prefix the attribute name with a dash ( - )
            to sort responses in descending order, (for example, '-name').
            Supported options are listed in ListComparisonChatsSortQueryParams,
            but the values can differ depending on platform version.
            The default sort parameter is None, which results in
            comparison chats returning in order of creation time, descending.

        Returns
        -------
        comparison_chats : list[ComparisonChat]
            Returns a list of comparison chats.
        """
        params = {
            "playground_id": get_entity_id(playground) if playground else None,
            "sort": sort,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def delete(self) -> None:
        """
        Delete the single comparison chat.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)

    def update(self, name: str) -> ComparisonChat:
        """
        Update the comparison chat.

        Parameters
        ----------
        name : str
            The new name for the comparison chat.

        Returns
        -------
        comparison_chat : ComparisonChat
            The updated comparison chat.
        """
        payload = {"name": name}
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        r_data = self._client.patch(url, data=payload)
        return self.from_server_data(r_data.json())



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import cast, Dict, List, Union

import trafaret as t

from datarobot.insights.base import BaseInsight


class RocCurve(BaseInsight):
    """Class for ROC Curve calculations. Use the standard methods of BaseInsight to compute
    and retrieve: compute, create, list, get.

    Usage example:

        ```python
        >>> from datarobot.insights import RocCurve
        >>> RocCurve.compute("67643b2d87bb4954d7917323", data_slice_id="6764389b4bdd48581485a58b")
        <datarobot.models.status_check_job.StatusCheckJob object at 0x7fc0b313cc40>
        >>> RocCurve.get("67643b2d87bb4954d7917323", data_slice_id="6764389b4bdd48581485a58b")
        <datarobot.insights.roc_curve.RocCurve object at 0x7fc0a7549d20>
        >>> RocCurve.list("67643b2d87bb4954d7917323")
        [<datarobot.insights.roc_curve.RocCurve object at 0x7fc0a7549f60>, ...]
        >>> RocCurve.list("67643b2d87bb4954d7917323")[0].roc_points
        [{'accuracy': 0.539375, 'f1_score': 0.0, 'false_negative_score': 737, 'true_negative_score': 863, ...}]
        ```
    """

    ROC_POINTS = t.Dict(
        {
            t.Key("accuracy"): t.Float(),
            t.Key("f1Score"): t.Float(),
            t.Key("falseNegativeScore"): t.Int(),
            t.Key("trueNegativeScore"): t.Int(),
            t.Key("falsePositiveScore"): t.Int(),
            t.Key("truePositiveScore"): t.Int(),
            t.Key("trueNegativeRate"): t.Float(),
            t.Key("falsePositiveRate"): t.Float(),
            t.Key("truePositiveRate"): t.Float(),
            t.Key("matthewsCorrelationCoefficient"): t.Float(),
            t.Key("positivePredictiveValue"): t.Float(),
            t.Key("negativePredictiveValue"): t.Float(),
            t.Key("threshold"): t.Float(),
            t.Key("fractionPredictedAsPositive"): t.Float(),
            t.Key("fractionPredictedAsNegative"): t.Float(),
            t.Key("liftPositive"): t.Float(),
            t.Key("liftNegative"): t.Float(),
        }
    ).ignore_extra("*")

    INSIGHT_NAME = "rocCurve"
    INSIGHT_DATA = {
        t.Key("rocPoints"): t.List(ROC_POINTS),
        t.Key("negativeClassPredictions"): t.List(t.Float()),
        t.Key("positiveClassPredictions"): t.List(t.Float()),
        t.Key("auc", optional=True): t.Float(),
        t.Key("kolmogorovSmirnovMetric", optional=True): t.Float(),
    }

    @property
    def kolmogorov_smirnov_metric(self) -> float:
        """Kolmogorov-Smirnov metric for the ROC curve values"""
        return cast(float, self.data["kolmogorov_smirnov_metric"])

    @property
    def auc(self) -> float:
        """AUC metric for the ROC curve values"""
        return cast(float, self.data["auc"])

    @property
    def positive_class_predictions(self) -> List[float]:
        """List of positive class prediction values for the ROC curve"""
        return cast(List[float], self.data["positive_class_predictions"])

    @property
    def negative_class_predictions(self) -> List[float]:
        """List of negative class prediction values for the ROC curve"""
        return cast(List[float], self.data["negative_class_predictions"])

    @property
    def roc_points(self) -> List[Dict[str, Union[int, float]]]:
        """List of ROC values for the ROC curve"""
        return cast(List[Dict[str, Union[int, float]]], self.data["roc_points"])



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from urllib.parse import quote

import pandas as pd
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.models.api_object import APIObject
from datarobot.models.pairwise_statistics import (
    PairwiseConditionalProbabilities,
    PairwiseCorrelations,
    PairwiseJointProbabilities,
)
from datarobot.utils.waiters import wait_for_async_resolution

from ..errors import InvalidUsageError


class HasHistogram:  # pylint: disable=missing-class-docstring
    def get_histogram(self, bin_limit=None):
        """Retrieve a feature histogram

        Parameters
        ----------
        bin_limit : int or None
            Desired max number of histogram bins. If omitted, by default
            endpoint will use 60.

        Returns
        -------
        featureHistogram : FeatureHistogram
            The requested histogram with desired number or bins
        """
        return FeatureHistogram.get(self.project_id, self.name, bin_limit)


class Feature(APIObject, HasHistogram):
    """A feature from a project's dataset

    These are features either included in the originally uploaded dataset or added to it via
    feature transformations.  In time series projects, these will be distinct from the
    :class:`ModelingFeature <datarobot.models.ModelingFeature>` s created during partitioning;
    otherwise, they will correspond to the same features.  For more information about input and
    modeling features, see the :ref:`time series documentation<input-vs-modeling>`.

    The ``min``, ``max``, ``mean``, ``median``, and ``std_dev`` attributes provide information about
    the distribution of the feature in the EDA sample data.  For non-numeric features or features
    created prior to these summary statistics becoming available, they will be None.  For features
    where the summary statistics are available, they will be in a format compatible with the data
    type, i.e. date type features will have their summary statistics expressed as ISO-8601
    formatted date strings.

    Attributes
    ----------
    id : int
        the id for the feature - note that `name` is used to reference the feature instead of `id`
    project_id : str
        the id of the project the feature belongs to
    name : str
        the name of the feature
    feature_type : str
        the type of the feature, e.g. 'Categorical', 'Text'
    importance : float or None
        numeric measure of the strength of relationship between the feature and target (independent
        of any model or other features); may be None for non-modeling features such as partition
        columns
    low_information : bool
        whether a feature is considered too uninformative for modeling (e.g. because it has too few
        values)
    unique_count : int
        number of unique values
    na_count : int or None
        number of missing values
    date_format : str or None
        For Date features, the date format string for how this feature
        was interpreted, compatible with https://docs.python.org/2/library/time.html#time.strftime .
        For other feature types, None.
    min : str, int, float, or None
        The minimum value of the source data in the EDA sample
    max : str, int, float, or None
        The maximum value of the source data in the EDA sample
    mean : str, int, or, float
        The arithmetic mean of the source data in the EDA sample
    median : str, int, float, or None
        The median of the source data in the EDA sample
    std_dev : str, int, float, or None
        The standard deviation of the source data in the EDA sample
    time_series_eligible : bool
        Whether this feature can be used as the datetime partition column in a time series project.
    time_series_eligibility_reason : str
        Why the feature is ineligible for the datetime partition column in a time series project,
        or 'suitable' when it is eligible.
    time_step : int or None
        For time series eligible features, a positive integer determining the interval at which
        windows can be specified. If used as the datetime partition column on a time series
        project, the feature derivation and forecast windows must start and end at an integer
        multiple of this value. None for features that are not time series eligible.
    time_unit : str or None
        For time series eligible features, the time unit covered by a single time step, e.g. 'HOUR',
        or None for features that are not time series eligible.
    target_leakage : str
        Whether a feature is considered to have target leakage or not.  A value of
        'SKIPPED_DETECTION' indicates that target leakage detection was not run on the feature.
        'FALSE' indicates no leakage, 'MODERATE' indicates a moderate risk of target leakage, and
        'HIGH_RISK' indicates a high risk of target leakage
    feature_lineage_id : str
        id of a lineage for automatically discovered features or derived time series features.
    key_summary: list of dict
        Statistics for top 50 keys (truncated to 103 characters) of
        Summarized Categorical column example:

        {\'key\':\'DataRobot\',
        \'summary\':{\'min\':0, \'max\':29815.0, \'stdDev\':6498.029, \'mean\':1490.75,
        \'median\':0.0, \'pctRows\':5.0}}

        where,
            key: string or None
                name of the key
            summary: dict
                statistics of the key

                max: maximum value of the key.
                min: minimum value of the key.
                mean: mean value of the key.
                median: median value of the key.
                stdDev: standard deviation of the key.
                pctRows: percentage occurrence of key in the EDA sample of the feature.
    multilabel_insights_key : str or None
        For multicategorical columns this will contain a key for multilabel insights. The key is
        unique for a project, feature and EDA stage combination. This will be the key for the most
        recent, finished EDA stage.
    """

    _converter = t.Dict(
        {
            t.Key("id"): Int,
            t.Key("project_id"): String,
            t.Key("name"): String,
            t.Key("feature_type", optional=True): String,
            t.Key("importance", optional=True): t.Float,
            t.Key("low_information"): t.Bool,
            t.Key("unique_count"): Int,
            t.Key("na_count", optional=True): Int,
            t.Key("date_format", optional=True): String,
            t.Key("min", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("max", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("mean", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("median", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("std_dev", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("time_series_eligible"): t.Bool,
            t.Key("time_series_eligibility_reason"): String,
            t.Key("time_step", optional=True): Int,
            t.Key("time_unit", optional=True): String,
            t.Key("target_leakage", optional=True): String,
            t.Key("feature_lineage_id", optional=True): t.Or(String, t.Null),
            t.Key("key_summary", optional=True): t.List(
                t.Dict(
                    {
                        t.Key("key"): String(allow_blank=True),
                        t.Key("summary"): t.Dict(
                            {
                                t.Key("max"): t.Or(t.Float, Int),
                                t.Key("min"): t.Or(t.Float, Int),
                                t.Key("mean"): t.Or(t.Float, Int),
                                t.Key("median"): t.Or(t.Float, Int),
                                t.Key("std_dev"): t.Or(t.Float, Int),
                                t.Key("pct_rows"): t.Or(t.Float, Int),
                            }
                        ).allow_extra("*"),
                    }
                ).allow_extra("*")
            ),
            t.Key("multilabel_insights", optional=True): t.Dict(
                {t.Key("multilabel_insights_key"): String}
            ).allow_extra("*"),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id,
        project_id=None,
        name=None,
        feature_type=None,
        importance=None,
        low_information=None,
        unique_count=None,
        na_count=None,
        date_format=None,
        min=None,
        max=None,
        mean=None,
        median=None,
        std_dev=None,
        time_series_eligible=None,
        time_series_eligibility_reason=None,
        time_step=None,
        time_unit=None,
        target_leakage=None,
        feature_lineage_id=None,
        key_summary=None,
        multilabel_insights=None,
    ):
        self.id = id
        self.project_id = project_id
        self.name = name
        self.feature_type = feature_type
        self.importance = importance
        self.low_information = low_information
        self.unique_count = unique_count
        self.na_count = na_count
        self.date_format = date_format
        self.min = min
        self.max = max
        self.mean = mean
        self.median = median
        self.std_dev = std_dev
        self.time_series_eligible = time_series_eligible
        self.time_series_eligibility_reason = time_series_eligibility_reason
        self.time_step = time_step
        self.time_unit = time_unit
        self.target_leakage = target_leakage
        self.feature_lineage_id = feature_lineage_id
        self.key_summary = key_summary
        self.multilabel_insights_key = (
            multilabel_insights["multilabel_insights_key"] if multilabel_insights else None
        )

    def __repr__(self):
        return f"Feature({self.name})"

    @classmethod
    def get(cls, project_id, feature_name):
        """Retrieve a single feature

        Parameters
        ----------
        project_id : str
            The ID of the project the feature is associated with.
        feature_name : str
            The name of the feature to retrieve

        Returns
        -------
        feature : Feature
            The queried instance
        """
        path = cls._build_url(project_id, feature_name)
        return cls.from_location(path)

    @classmethod
    def _build_url(cls, project_id, feature_name):
        return "projects/{}/features/{}/".format(
            project_id,
            feature_name if isinstance(feature_name, int) else quote(feature_name.encode("utf-8")),
        )

    def get_multiseries_properties(self, multiseries_id_columns, max_wait=DEFAULT_MAX_WAIT):
        """Retrieve time series properties for a potential multiseries datetime partition column

        Multiseries time series projects use multiseries id columns to model multiple distinct
        series within a single project.  This function returns the time series properties (time step
        and time unit) of this column if it were used as a datetime partition column with the
        specified multiseries id columns, running multiseries detection automatically if it had not
        previously been successfully ran.

        Parameters
        ----------
        multiseries_id_columns : List[str]
            the name(s) of the multiseries id columns to use with this datetime partition column.
            Currently only one multiseries id column is supported.
        max_wait : Optional[int]
            if a multiseries detection task is run, the maximum amount of time to wait for it to
            complete before giving up

        Returns
        -------
        properties : dict
            A dict with three keys:

                - time_series_eligible : bool, whether the column can be used as a partition column
                - time_unit : str or null, the inferred time unit if used as a partition column
                - time_step : int or null, the inferred time step if used as a partition column
        """

        def _extract_properties():
            retrieve_url = "{}multiseriesProperties/".format(
                self._build_url(self.project_id, self.name)
            )
            response = self._client.get(retrieve_url)
            response_schema = t.Dict(
                {
                    t.Key("datetimePartitionColumn"): String(),
                    t.Key("detectedMultiseriesIdColumns"): t.List(
                        t.Dict(
                            {
                                t.Key("multiseriesIdColumns"): t.List(String()),
                                t.Key("timeUnit"): String(),
                                t.Key("timeStep"): Int(),
                            }
                        ).ignore_extra("*")
                    ),
                }
            ).ignore_extra("*")
            response_data = response_schema.check(response.json())
            detected_columns = response_data["detectedMultiseriesIdColumns"]

            detected = next(
                (
                    col_set
                    for col_set in detected_columns
                    if set(col_set["multiseriesIdColumns"]) == set(multiseries_id_columns)
                ),
                None,
            )
            if detected is not None:
                return (
                    True,
                    {
                        "time_series_eligible": True,
                        "time_unit": detected["timeUnit"],
                        "time_step": detected["timeStep"],
                    },
                )
            return False, {"time_series_eligible": False, "time_unit": None, "time_step": None}

        detect_url = f"projects/{self.project_id}/multiseriesProperties/"
        was_detected, properties = _extract_properties()
        if was_detected:
            return properties

        detection_job = self._client.post(
            detect_url,
            data={
                "datetimePartitionColumn": self.name,
                "multiseriesIdColumns": multiseries_id_columns,
            },
        ).headers["Location"]
        wait_for_async_resolution(self._client, detection_job, max_wait=max_wait)
        return _extract_properties()[1]

    def get_cross_series_properties(
        self, datetime_partition_column, cross_series_group_by_columns, max_wait=DEFAULT_MAX_WAIT
    ):
        """Retrieve cross-series properties for multiseries ID column.

        This function returns the cross-series properties (eligibility
        as group-by column) of this column if it were used with specified datetime partition column
        and with current multiseries id column, running cross-series group-by validation
        automatically if it had not previously been successfully ran.

        Parameters
        ----------
        datetime_partition_column : datetime partition column
        cross_series_group_by_columns : List[str]
            the name(s) of the columns to use with this multiseries ID column.
            Currently only one cross-series group-by column is supported.
        max_wait : Optional[int]
            if a multiseries detection task is run, the maximum amount of time to wait for it to
            complete before giving up

        Returns
        -------
        properties : dict
            A dict with three keys:

                - name : str, column name
                - eligibility : str, reason for column eligibility
                - isEligible : bool, is column eligible as cross-series group-by
        """

        def _extract_properties():
            retrieve_url = (
                "projects/{}/multiseriesIds/{}/crossSeriesProperties/?"
                "crossSeriesGroupByColumns={}"
            ).format(self.project_id, self.name, cross_series_group_by_columns[0])

            response = self._client.get(retrieve_url)
            response_schema = t.Dict(
                {
                    t.Key("multiseriesId"): String(),
                    t.Key("crossSeriesGroupByColumns"): t.List(
                        t.Dict(
                            {
                                t.Key("name"): String(),
                                t.Key("eligibility"): String(),
                                t.Key("isEligible"): t.Bool(),
                            }
                        ).ignore_extra("*")
                    ),
                }
            ).ignore_extra("*")

            response_data = response_schema.check(response.json())
            for col in response_data["crossSeriesGroupByColumns"]:
                if col["name"] == cross_series_group_by_columns[0]:
                    if col["eligibility"] == "notAnalyzed":
                        return False, col
                    else:
                        return True, col
            return (
                False,
                {
                    "eligibility": "notAnalyzed",
                    "isEligible": False,
                    "name": cross_series_group_by_columns[0],
                },
            )

        validation_url = f"projects/{self.project_id}/crossSeriesProperties/"
        was_detected, properties = _extract_properties()
        if was_detected:
            return properties

        req_data = {
            "datetimePartitionColumn": datetime_partition_column,
            "multiseriesIdColumn": self.name,
            "crossSeriesGroupByColumns": cross_series_group_by_columns,
        }
        detection_job = self._client.post(validation_url, data=req_data).headers["Location"]
        wait_for_async_resolution(self._client, detection_job, max_wait=max_wait)
        return _extract_properties()[1]

    def get_multicategorical_histogram(self):
        """Retrieve multicategorical histogram for this feature

        .. versionadded:: v2.24

        Returns
        -------
        :class:`datarobot.models.MulticategoricalHistogram`

        Raises
        ------
        datarobot.errors.InvalidUsageError
            if this method is called on a unsuited feature
        ValueError
            if no multilabel_insights_key is present for this feature
        """
        if self.multilabel_insights_key:
            return MulticategoricalHistogram.get(self.multilabel_insights_key)
        elif self.feature_type != "Multicategorical":
            raise InvalidUsageError(
                "Multicategorical Histograms are only available for features"
                "with feature_type Multicategorical. "
                "This feature is of type: {}".format(self.feature_type)
            )
        else:
            raise ValueError("A valid multilabel_insights_key is required, but not present")

    def get_pairwise_correlations(self):
        """Retrieve pairwise label correlation for multicategorical features

        .. versionadded:: v2.24

        Returns
        -------
        :class:`datarobot.models.PairwiseCorrelations`

        Raises
        ------
        datarobot.errors.InvalidUsageError
            if this method is called on a unsuited feature
        ValueError
            if no multilabel_insights_key is present for this feature
        """
        if self.multilabel_insights_key:
            return PairwiseCorrelations.get(self.multilabel_insights_key)
        elif self.feature_type != "Multicategorical":
            raise InvalidUsageError(
                "Pairwise Correlation is only available for features"
                "with feature_type Multicategorical. "
                "This feature is of type: {}".format(self.feature_type)
            )
        else:
            raise ValueError("A valid multilabel_insights_key is required, but not present")

    def get_pairwise_joint_probabilities(self):
        """Retrieve pairwise label joint probabilities for multicategorical features

        .. versionadded:: v2.24

        Returns
        -------
        :class:`datarobot.models.PairwiseJointProbabilities`

        Raises
        ------
        datarobot.errors.InvalidUsageError
            if this method is called on a unsuited feature
        ValueError
            if no multilabel_insights_key is present for this feature
        """
        if self.multilabel_insights_key:
            return PairwiseJointProbabilities.get(self.multilabel_insights_key)
        elif self.feature_type != "Multicategorical":
            raise InvalidUsageError(
                "Pairwise Joint Probability is only available for features"
                "with feature_type Multicategorical. "
                "This feature is of type: {}".format(self.feature_type)
            )
        else:
            raise ValueError("A valid multilabel_insights_key is required, but not present")

    def get_pairwise_conditional_probabilities(self):
        """Retrieve pairwise label conditional probabilities for multicategorical features

        .. versionadded:: v2.24

        Returns
        -------
        :class:`datarobot.models.PairwiseConditionalProbabilities`

        Raises
        ------
        datarobot.errors.InvalidUsageError
            if this method is called on a unsuited feature
        ValueError
            if no multilabel_insights_key is present for this feature
        """
        if self.multilabel_insights_key:
            return PairwiseConditionalProbabilities.get(self.multilabel_insights_key)
        elif self.feature_type != "Multicategorical":
            raise InvalidUsageError(
                "Pairwise Conditional Probability is only available for features"
                "with feature_type Multicategorical. "
                "This feature is of type: {}".format(self.feature_type)
            )
        else:
            raise ValueError("A valid multilabel_insights_key is required, but not present")


class ModelingFeature(APIObject, HasHistogram):
    """A feature used for modeling

    In time series projects, a new set of modeling features is created after setting the
    partitioning options.  These features are automatically derived from those in the project's
    dataset and are the features used for modeling.  Modeling features are only accessible once
    the target and partitioning options have been set.  In projects that don't use time series
    modeling, once the target has been set, ModelingFeatures and Features will behave
    the same.

    For more information about input and modeling features, see the
    :ref:`time series documentation<input-vs-modeling>`.

    As with the :class:`Feature <datarobot.models.Feature>` object, the `min`, `max, `mean`,
    `median`, and `std_dev` attributes provide information about the distribution of the feature in
    the EDA sample data.  For non-numeric features, they will be None.  For features where the
    summary statistics are available, they will be in a format compatible with the data type, i.e.
    date type features will have their summary statistics expressed as ISO-8601 formatted date
    strings.

    Attributes
    ----------
    project_id : str
        the id of the project the feature belongs to
    name : str
        the name of the feature
    feature_type : str
        the type of the feature, e.g. 'Categorical', 'Text'
    importance : float or None
        numeric measure of the strength of relationship between the feature and target (independent
        of any model or other features); may be None for non-modeling features such as partition
        columns
    low_information : bool
        whether a feature is considered too uninformative for modeling (e.g. because it has too few
        values)
    unique_count : int
        number of unique values
    na_count : int or None
        number of missing values
    date_format : str or None
        For Date features, the date format string for how this feature
        was interpreted, compatible with https://docs.python.org/2/library/time.html#time.strftime .
        For other feature types, None.
    min : str, int, float, or None
        The minimum value of the source data in the EDA sample
    max : str, int, float, or None
        The maximum value of the source data in the EDA sample
    mean : str, int, or, float
        The arithmetic mean of the source data in the EDA sample
    median : str, int, float, or None
        The median of the source data in the EDA sample
    std_dev : str, int, float, or None
        The standard deviation of the source data in the EDA sample
    parent_feature_names : List[str]
        A list of the names of input features used to derive this modeling feature.  In cases where
        the input features and modeling features are the same, this will simply contain the
        feature's name.  Note that if a derived feature was used to create this modeling feature,
        the values here will not necessarily correspond to the features that must be supplied at
        prediction time.
    key_summary: list of dict
        Statistics for top 50 keys (truncated to 103 characters) of
        Summarized Categorical column example:

        {\'key\':\'DataRobot\',
        \'summary\':{\'min\':0, \'max\':29815.0, \'stdDev\':6498.029, \'mean\':1490.75,
        \'median\':0.0, \'pctRows\':5.0}}

        where,
            key: string or None
                name of the key
            summary: dict
                statistics of the key

                max: maximum value of the key.
                min: minimum value of the key.
                mean: mean value of the key.
                median: median value of the key.
                stdDev: standard deviation of the key.
                pctRows: percentage occurrence of key in the EDA sample of the feature.
    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("name"): String,
            t.Key("feature_type", optional=True): String,
            t.Key("importance", optional=True): t.Float,
            t.Key("low_information"): t.Bool,
            t.Key("unique_count"): Int,
            t.Key("na_count", optional=True): Int,
            t.Key("date_format", optional=True): String,
            t.Key("min", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("max", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("mean", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("median", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("std_dev", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("parent_feature_names"): t.List(String),
            t.Key("is_restored_after_reduction", optional=True): t.Bool,
            t.Key("key_summary", optional=True): t.List(
                t.Dict(
                    {
                        t.Key("key"): String(allow_blank=True),
                        t.Key("summary"): t.Dict(
                            {
                                t.Key("max"): t.Or(t.Float, Int),
                                t.Key("min"): t.Or(t.Float, Int),
                                t.Key("mean"): t.Or(t.Float, Int),
                                t.Key("median"): t.Or(t.Float, Int),
                                t.Key("std_dev"): t.Or(t.Float, Int),
                                t.Key("pct_rows"): t.Or(t.Float, Int),
                            }
                        ).allow_extra("*"),
                    }
                ).allow_extra("*")
            ),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id=None,
        name=None,
        feature_type=None,
        importance=None,
        low_information=None,
        unique_count=None,
        na_count=None,
        date_format=None,
        min=None,
        max=None,
        mean=None,
        median=None,
        std_dev=None,
        parent_feature_names=None,
        key_summary=None,
        is_restored_after_reduction=None,
    ):
        self.project_id = project_id
        self.name = name
        self.feature_type = feature_type
        self.importance = importance
        self.low_information = low_information
        self.unique_count = unique_count
        self.na_count = na_count
        self.date_format = date_format
        self.min = min
        self.max = max
        self.mean = mean
        self.median = median
        self.std_dev = std_dev
        self.parent_feature_names = parent_feature_names
        self.key_summary = key_summary
        self.is_restored_after_reduction = is_restored_after_reduction

    def __repr__(self):
        return f"ModelingFeature({self.name})"

    @classmethod
    def get(cls, project_id, feature_name):
        """Retrieve a single modeling feature

        Parameters
        ----------
        project_id : str
            The ID of the project the feature is associated with.
        feature_name : str
            The name of the feature to retrieve

        Returns
        -------
        feature : ModelingFeature
            The requested feature
        """
        path = cls._build_url(project_id, feature_name)
        return cls.from_location(path)

    @classmethod
    def _build_url(cls, project_id, feature_name):
        return "projects/{}/modelingFeatures/{}/".format(
            project_id, quote(feature_name.encode("utf-8"))
        )


class DatasetFeature(APIObject):
    """A feature from a project's dataset

    These are features either included in the originally uploaded dataset or added to it via
    feature transformations.

    The ``min``, ``max``, ``mean``, ``median``, and ``std_dev`` attributes provide information about
    the distribution of the feature in the EDA sample data.  For non-numeric features or features
    created prior to these summary statistics becoming available, they will be None.  For features
    where the summary statistics are available, they will be in a format compatible with the data
    type, i.e. date type features will have their summary statistics expressed as ISO-8601
    formatted date strings.

    Attributes
    ----------
    id : int
        the id for the feature - note that `name` is used to reference the feature instead of `id`
    dataset_id : str
        the id of the dataset the feature belongs to
    dataset_version_id : str
        the id of the dataset version the feature belongs to
    name : str
        the name of the feature
    feature_type : Optional[str]
        the type of the feature, e.g. 'Categorical', 'Text'
    low_information : Optional[bool]
        whether a feature is considered too uninformative for modeling (e.g. because it has too few
        values)
    unique_count : Optional[int]
        number of unique values
    na_count : Optional[int]
        number of missing values
    date_format : Optional[str]
        For Date features, the date format string for how this feature
        was interpreted, compatible with https://docs.python.org/2/library/time.html#time.strftime .
        For other feature types, None.
    min : str, int, Optional[float]
        The minimum value of the source data in the EDA sample
    max : str, int, Optional[float]
        The maximum value of the source data in the EDA sample
    mean : str, int, Optional[float]
        The arithmetic mean of the source data in the EDA sample
    median : str, int, Optional[float]
        The median of the source data in the EDA sample
    std_dev : str, int, Optional[float]
        The standard deviation of the source data in the EDA sample
    time_series_eligible : Optional[bool]
        Whether this feature can be used as the datetime partition column in a time series project.
    time_series_eligibility_reason : Optional[str]
        Why the feature is ineligible for the datetime partition column in a time series project,
        or 'suitable' when it is eligible.
    time_step : Optional[int]
        For time series eligible features, a positive integer determining the interval at which
        windows can be specified. If used as the datetime partition column on a time series
        project, the feature derivation and forecast windows must start and end at an integer
        multiple of this value. None for features that are not time series eligible.
    time_unit : Optional[str]
        For time series eligible features, the time unit covered by a single time step, e.g. 'HOUR',
        or None for features that are not time series eligible.
    target_leakage : Optional[str]
        Whether a feature is considered to have target leakage or not.  A value of
        'SKIPPED_DETECTION' indicates that target leakage detection was not run on the feature.
        'FALSE' indicates no leakage, 'MODERATE' indicates a moderate risk of target leakage, and
        'HIGH_RISK' indicates a high risk of target leakage
    target_leakage_reason: string, optional
        The descriptive text explaining the reason for target leakage, if any.
    """

    _converter = t.Dict(
        {
            t.Key("id") >> "id_": Int,
            t.Key("dataset_id"): String,
            t.Key("dataset_version_id"): String,
            t.Key("name"): String,
            t.Key("feature_type", optional=True): String,
            t.Key("low_information", optional=True): t.Bool,
            t.Key("unique_count", optional=True): Int,
            t.Key("na_count", optional=True): Int,
            t.Key("date_format", optional=True): String,
            t.Key("min", optional=True) >> "min_": t.Or(String, Int, t.Float, t.Null),
            t.Key("max", optional=True) >> "max_": t.Or(String, Int, t.Float, t.Null),
            t.Key("mean", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("median", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("std_dev", optional=True): t.Or(String, Int, t.Float, t.Null),
            t.Key("time_series_eligible", optional=True): t.Bool,
            t.Key("time_series_eligibility_reason", optional=True): String,
            t.Key("time_step", optional=True): Int,
            t.Key("time_unit", optional=True): String,
            t.Key("target_leakage", optional=True): String,
            t.Key("target_leakage_reason", optional=True): String,
        }
    ).allow_extra("*")

    def __init__(
        self,
        id_,
        dataset_id=None,
        dataset_version_id=None,
        name=None,
        feature_type=None,
        low_information=None,
        unique_count=None,
        na_count=None,
        date_format=None,
        min_=None,
        max_=None,
        mean=None,
        median=None,
        std_dev=None,
        time_series_eligible=None,
        time_series_eligibility_reason=None,
        time_step=None,
        time_unit=None,
        target_leakage=None,
        target_leakage_reason=None,
    ):
        self.id = id_
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id
        self.name = name
        self.feature_type = feature_type
        self.low_information = low_information
        self.unique_count = unique_count
        self.na_count = na_count
        self.date_format = date_format
        self.min = min_
        self.max = max_
        self.mean = mean
        self.median = median
        self.std_dev = std_dev
        self.time_series_eligible = time_series_eligible
        self.time_series_eligibility_reason = time_series_eligibility_reason
        self.time_step = time_step
        self.time_unit = time_unit
        self.target_leakage = target_leakage
        self.target_leakage_reason = target_leakage_reason

    def __repr__(self):
        return f"Feature({self.name}, dataset_id={self.dataset_id})"

    def get_histogram(self, bin_limit=None):
        """Retrieve a feature histogram

        Parameters
        ----------
        bin_limit : int or None
            Desired max number of histogram bins. If omitted, by default
            endpoint will use 60.

        Returns
        -------
        featureHistogram : DatasetFeatureHistogram
            The requested histogram with desired number or bins
        """
        return DatasetFeatureHistogram.get(self.dataset_id, self.name, bin_limit)


class BaseFeatureHistogram(APIObject):
    """A histogram plot data for a specific feature

    .. versionadded:: v2.14

    Histogram is a popular way of visual representation of feature values
    distribution. Here histogram is represented as an ordered collection of bins.
    For categorical features every bin represents exactly one of feature values
    and the count in that bin is the number of occurrences of that value.
    For numeric features every bin represents a range of values (low end inclusive,
    high end exclusive) and the count in the bin is the total number of occurrences of all
    values in this range. In addition, each bin may contain a target feature average for values
    in that bin (see ``target`` description below).

    Notes
    -----

    ``HistogramBin`` contains:

    * ``label`` : (str) for categorical features: the value of the feature,
      for numeric: the low end of bin range,
      so that the difference between two consecutive bin labels is the length
      of the bin
    * ``count`` : (int or float) number of values in this bin's range
      If project uses weights, the value is equal to the sum of weights of
      all feature values in bin's range
    * ``target`` : (float or None) Average of the target feature values for the bin.
      Present only for informative features if project target has already been selected and
      AIM processing has finished. For multiclass projects the value is always null.

    Attributes
    ----------
    plot : list
        a list of dictionaries with a schema described as ``HistogramBin``

    """

    _converter = t.Dict(
        {
            t.Key("plot"): t.List(
                t.Dict(
                    {
                        t.Key("label"): String,
                        t.Key("count"): t.Or(Int, t.Float),
                        t.Key("target", default=None): t.Or(t.Float, t.Null),
                    }
                )
            )
        }
    )

    _root_path = None

    def __init__(self, plot):
        self.plot = plot

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @classmethod
    def get(cls, object_id, feature_name, bin_limit=None, key_name=None):
        """Retrieve a single feature histogram

        Parameters
        ----------
        object_id : str
            The ID of the object the feature is associated with.
        feature_name : str
            The name of the feature to retrieve
        bin_limit : int or None
            Desired max number of histogram bins. If omitted, by default
            endpoint will use 60.
        key_name: string or None
            (Only required for summarized categorical feature)
            Name of the top 50 keys for which plot to be retrieved

        Returns
        -------
        featureHistogram : FeatureHistogram
            The queried instance with `plot` attribute in it.
        """
        path = cls._build_url(object_id, feature_name, bin_limit, key_name)
        return cls.from_location(path)

    @classmethod
    def _build_url(
        cls, project_id, feature_name, bin_limit=None, key_name=None
    ):  # pylint: disable=missing-function-docstring
        url = "{}/{}/featureHistograms/{}/".format(
            cls._root_path, project_id, quote(feature_name.encode("utf-8"))
        )

        if bin_limit is not None:  # makes sure pass 0 to endpoint
            url = f"{url}?binLimit={bin_limit}"
        if key_name is not None:
            url = "{}?key={}".format(url, quote(key_name.encode("utf-8")))
        return url


class FeatureHistogram(BaseFeatureHistogram):  # pylint: disable=missing-class-docstring
    _root_path = "projects"

    @classmethod
    def get(  # pylint: disable=arguments-renamed
        cls,
        project_id,
        feature_name,
        bin_limit=None,
        key_name=None,
    ):
        """Retrieve a single feature histogram

        Parameters
        ----------
        project_id : str
            The ID of the project the feature is associated with.
        feature_name : str
            The name of the feature to retrieve
        bin_limit : int or None
            Desired max number of histogram bins. If omitted, by default
            endpoint will use 60.
        key_name: string or None
            (Only required for summarized categorical feature)
            Name of the top 50 keys for which plot to be retrieved

        Returns
        -------
        featureHistogram : FeatureHistogram
            The queried instance with `plot` attribute in it.
        """
        return super().get(project_id, feature_name, bin_limit, key_name)


class DatasetFeatureHistogram(BaseFeatureHistogram):  # pylint: disable=missing-class-docstring
    _root_path = "datasets"

    @classmethod
    def get(  # pylint: disable=arguments-renamed
        cls,
        dataset_id,
        feature_name,
        bin_limit=None,
        key_name=None,
    ):
        """Retrieve a single feature histogram

        Parameters
        ----------
        dataset_id : str
            The ID of the Dataset the feature is associated with.
        feature_name : str
            The name of the feature to retrieve
        bin_limit : int or None
            Desired max number of histogram bins. If omitted, by default
            the endpoint will use 60.
        key_name: string or None
            (Only required for summarized categorical feature)
            Name of the top 50 keys for which plot to be retrieved

        Returns
        -------
        featureHistogram : FeatureHistogram
            The queried instance with `plot` attribute in it.
        """
        return super().get(dataset_id, feature_name, bin_limit, key_name)


class InteractionFeature(APIObject):
    """
    Interaction feature data

    .. versionadded:: v2.21

    Attributes
    ----------
    rows: int
        Total number of rows
    source_columns: list(str)
        names of two categorical features which were combined into this one
    bars: list(dict)
        dictionaries representing frequencies of each independent value from the source columns
    bubbles: list(dict)
        dictionaries representing frequencies of each combined value in the interaction feature.
    """

    _converter = t.Dict(
        {
            t.Key("rows"): Int,
            t.Key("source_columns"): t.List(String, min_length=2, max_length=2),
            t.Key("bars"): t.List(
                t.Dict(
                    {
                        t.Key("column_name"): String,
                        t.Key("counts"): t.List(
                            t.Dict({t.Key("value"): String, t.Key("count"): Int}),
                            min_length=2,
                            max_length=2,
                        ),
                    }
                ).allow_extra("*")
            ),
            t.Key("bubbles"): t.List(
                t.Dict(
                    {
                        t.Key("count"): Int,
                        t.Key("source_data"): t.List(
                            t.Dict({t.Key("value"): String, t.Key("column_name"): String}),
                            min_length=2,
                            max_length=2,
                        ),
                    }
                ).allow_extra("*")
            ),
        }
    ).allow_extra("*")

    def __init__(self, rows, source_columns, bars, bubbles):
        self.rows = rows
        self.bars = bars
        self.bubbles = bubbles
        self.source_columns = source_columns

    @classmethod
    def _url(cls, project_id, feature_name):
        safe_feature_name = quote(feature_name.encode("utf-8"))
        return f"projects/{project_id}/interactionFeatures/{safe_feature_name}/"

    @classmethod
    def get(cls, project_id, feature_name):
        """
        Retrieve a single Interaction feature

        Parameters
        ----------
        project_id : str
            The id of the project the feature belongs to
        feature_name : str
            The name of the Interaction feature to retrieve

        Returns
        -------
        feature : InteractionFeature
            The queried instance
        """
        feature_url = cls._url(project_id, feature_name)
        return cls.from_location(feature_url)


class FeatureLineage(APIObject):
    """Lineage of an automatically engineered feature.

    Attributes
    ----------
    steps: list
        list of steps which were applied to build the feature.

        ``steps`` structure is:

            id - (int)
                step id starting with 0.
            step_type: (str)
                one of the data/action/json/generatedData.
            name: (str)
                name of the step.
            description: (str)
                description of the step.
            parents: (list[int])
                references to other steps id.
            is_time_aware: (bool)
                indicator of step being time aware. Mandatory only for *action* and *join* steps.
                *action* step provides additional information about feature derivation window
                in the `timeInfo` field.
            catalog_id: (str)
                id of the catalog for a *data* step.
            catalog_version_id: (str)
                id of the catalog version for a *data* step.
            group_by: (list[str])
                list of columns which this *action* step aggregated by.
            columns: (list)
                names of columns involved into the feature generation. Available only for *data* steps.
            time_info: (dict)
                description of the feature derivation window which was applied to this *action* step.
            join_info: (list[dict])
                *join* step details.

        ``columns`` structure is

            data_type: (str)
                the type of the feature, e.g. 'Categorical', 'Text'
            is_input: (bool)
                indicates features which provided data to transform in this lineage.
            name: (str)
                feature name.
            is_cutoff: (bool)
                indicates a cutoff column.

        ``time_info`` structure is:

            latest: (dict)
                end of the feature derivation window applied.
            duration: (dict)
                size of the feature derivation window applied.

        ``latest`` and `duration` structure is:

            time_unit: (str)
                time unit name like 'MINUTE', 'DAY', 'MONTH' etc.
            duration: (int)
                value/size of this duration object.

        ``join_info`` structure is:

            join_type - (str)
                kind of join, left/right.
            left_table - (dict)
                information about a dataset which was considered as left.
            right_table - (str)
                information about a dataset which was considered as right.

        ``left_table`` and ``right_table`` structure is:

            `columns` - (list[str])
                list of columns which datasets were joined by.
            `datasteps` - (list[int])
                list of *data* steps id which brought the *columns* into the current step dataset.
    """

    _join_table_trafaret = t.Dict(
        {t.Key("datasteps"): t.List(Int(gte=1)), t.Key("columns"): t.List(String)}
    )

    _duration_trafaret = t.Dict(
        {
            t.Key("duration"): Int,
            t.Key("time_unit"): String,
            t.Key("is_original", optional=True): t.Or(t.Bool, t.Null),
        }
    )

    _columns_trafaret = t.Dict(
        {
            t.Key("name"): String,
            t.Key("data_type"): String,
            t.Key("is_input"): t.Bool,
            t.Key("is_cutoff", optional=True): t.Bool,
        }
    ).allow_extra("*")

    _converter = t.Dict(
        {
            t.Key("steps"): t.List(
                t.Dict(
                    {
                        t.Key("id"): t.Or(Int(gte=0), t.Null),
                        t.Key("step_type"): t.Regexp(regexp="data|join|generatedColumn|action"),
                        t.Key("parents"): t.List(Int),
                        t.Key("name", optional=True): String,
                        t.Key("description", optional=True): t.Or(String, t.Null),
                        t.Key("data_type", optional=True): t.Or(String, t.Null),
                        t.Key("catalog_id", optional=True): t.Or(String, t.Null),
                        t.Key("catalog_version_id", optional=True): t.Or(String, t.Null),
                        t.Key("columns", optional=True): t.List(_columns_trafaret),
                        t.Key("is_time_aware", optional=True): t.Bool,
                        t.Key("join_info", optional=True): t.Dict(
                            {
                                t.Key("join_type"): t.Regexp(regexp="left|right"),
                                t.Key("left_table"): _join_table_trafaret,
                                t.Key("right_table"): _join_table_trafaret,
                            }
                        ).ignore_extra("*"),
                        t.Key("group_by", optional=True): t.Or(t.List(String), t.Null),
                        t.Key("time_info", optional=True): t.Or(
                            t.Or(
                                t.Dict(
                                    {
                                        t.Key("latest", optional=True): _duration_trafaret,
                                        t.Key("duration", optional=True): _duration_trafaret,
                                        t.Key("lag", optional=True): _duration_trafaret,
                                    }
                                ),
                                t.Null,
                            )
                        ),
                    }
                ).ignore_extra("*")
            ),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        steps=None,
    ):
        self.steps = steps

    @classmethod
    def get(cls, project_id, id):
        """Retrieve a single FeatureLineage.

        Parameters
        ----------
        project_id : str
            The id of the project the feature belongs to
        id : str
            id of a feature lineage to retrieve

        Returns
        -------
        lineage : FeatureLineage
            The queried instance
        """
        return cls.from_location(f"projects/{project_id}/featureLineages/{id}/")


class MulticategoricalHistogram(APIObject):
    """
    Histogram for Multicategorical feature.

    .. versionadded:: v2.24

    Notes
    -----

    ``HistogramValues`` contains:

    * ``values.[].label`` : string - Label name
    * ``values.[].plot`` : list - Histogram for label
    * ``values.[].plot.[].label_relevance`` : int - Label relevance value
    * ``values.[].plot.[].row_count`` : int - Row count where label has given relevance
    * ``values.[].plot.[].row_pct`` : float - Percentage of rows where label has given relevance

    Attributes
    ----------
    feature_name : str
        Name of the feature
    values : list(dict)
        List of Histogram values with a schema described as ``HistogramValues``

    """

    _converter = t.Dict(
        {
            t.Key("feature_name"): String,
            t.Key("histogram"): t.List(
                t.Dict(
                    {
                        t.Key("label"): String,
                        t.Key("plot"): t.List(
                            t.Dict(
                                {
                                    t.Key("label_relevance"): Int,
                                    t.Key("row_count"): Int,
                                    t.Key("row_pct"): t.Float,
                                }
                            ).allow_extra("*")
                        ),
                    }
                ).allow_extra("*")
            ),
        }
    ).allow_extra("*")

    def __init__(self, feature_name, histogram):
        self.values = histogram
        self.feature_name = feature_name

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @classmethod
    def get(cls, multilabel_insights_key):
        """Retrieves multicategorical histogram

        You might find it more convenient to use
        :meth:`Feature.get_multicategorical_histogram
        <datarobot.models.Feature.get_multicategorical_histogram>`
        instead.

        Parameters
        ----------
        multilabel_insights_key: string
            Key for multilabel insights, unique for a project, feature and EDA stage combination.
            The multilabel_insights_key can be retrieved via
            ``Feature.multilabel_insights_key``.

        Returns
        -------
        MulticategoricalHistogram
            The multicategorical histogram for multilabel_insights_key
        """
        url = f"multilabelInsights/{multilabel_insights_key}/histogram/"
        return cls.from_location(url)

    def to_dataframe(self):
        """
        Convenience method to get all the information from this multicategorical_histogram instance
        in form of a ``pandas.DataFrame``.

        Returns
        -------
        pandas.DataFrame
            Histogram information as a multicategorical_histogram. The dataframe will contain these
            columns: feature_name, label, label_relevance, row_count and row_pct
        """
        rows = []
        for label_values in self.values:
            label = label_values["label"]
            for plot_value in label_values["plot"]:
                row = {
                    "feature_name": self.feature_name,
                    "label": label,
                    "label_relevance": plot_value["label_relevance"],
                    "row_count": plot_value["row_count"],
                    "row_pct": plot_value["row_pct"],
                }
                rows.append(row)
        return pd.DataFrame(rows)



from typing import Any, Dict, Union

from datarobot.utils import camelize_obj


class DatarobotUIMixin:
    """A mixin exposes `_repr_mimebundle_` that is called by ipython kernel for cell results output.
    Expects class to define `_datarobot_ui_mime_type` for mapping UI component for results rendering and
    `_get_datarobot_ui_data` method that provides data for UI component
    """

    # pylint: disable-next=unused-argument
    def _repr_mimebundle_(  # type: ignore[no-untyped-def]
        self, include=None, exclude=None
    ) -> Union[None, Dict[str, Any]]:
        if not hasattr(self, "_datarobot_ui_mime_type") or not hasattr(
            self, "_get_datarobot_ui_data"
        ):
            return None
        return {self._datarobot_ui_mime_type: {"data": camelize_obj(self._get_datarobot_ui_data())}}



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Dict, List

from datarobot.models.api_object import APIObject
from datarobot.models.moderation.configuration import ModerationConfiguration
from datarobot.models.moderation.overall import OverallModerationConfig
from datarobot.utils import to_api


class ModelVersionUpdate(APIObject):
    """
    Implements the operation provided by "Save Configuration" in moderation UI.
    All guard configurations and overall config is saved to a new custom model version.
    """

    _path = "guardConfigurations/toNewCustomModelVersion/"

    @classmethod
    def new_custom_model_version_with_config(
        cls,
        custom_model_id: str,
        overall_config: OverallModerationConfig,
        configs: List[ModerationConfiguration],
    ) -> str:
        """
        Create a new custom model version with the provided moderation configuration
        Parameters
        ----------
        custom_model_id
        overall_config
        configs

        Returns
        -------
        ID of the new custom model version.

        """
        # API payload is a subset of guard configuration fields
        # unroll the list comprehension to make mypy happy
        fields_to_drop = ["id", "createdAt", "creatorId", "entityId", "entityType"]
        fields_to_drop_if_empty = [
            "openaiApiKey",
            "openaiApiBase",
            "openaiDeploymentId",
            "openaiCredential",
            "errorMessage",
        ]
        config_data: List[Dict[str, str]] = []
        # mypy really hates to_api(), so several suppressions are needed here
        for config in configs:
            config_dict = to_api(config)
            for f in fields_to_drop:
                config_dict.pop(f, None)  # type: ignore [union-attr, call-arg, arg-type]
            for f in fields_to_drop_if_empty:
                if not config_dict.get(f):  # type: ignore [union-attr]
                    config_dict.pop(f, None)  # type: ignore [union-attr, call-arg, arg-type]
            config_data.append(config_dict)  # type: ignore [arg-type]
        # and a subset of overall configuration fields.
        payload = {
            "customModelId": custom_model_id,
            "overallConfig": {
                "timeoutSec": overall_config.timeout_sec,
                "timeoutAction": overall_config.timeout_action.value,
            },
            "data": config_data,
        }
        response = cls._client.post(
            cls._path,
            data=payload,
        )
        custom_model_version_id: str = response.json().get("customModelVersionId")
        return custom_model_version_id
